"""
"""

from numbers import Real
from typing import Any, Iterable, List, NoReturn, Optional, Sequence, Tuple

import numpy as np
from torch import Tensor

from ..cfg import DEFAULTS
from ..utils.utils_signal_t import normalize as normalize_t
from .base import Augmenter

__all__ = [
    "RandomRenormalize",
]


class RandomRenormalize(Augmenter):
    """
    Randomly re-normalize the ECG tensor,
    using the Z-score normalization method.
    """

    __name__ = "RandomRenormalize"

    def __init__(
        self,
        mean: Iterable[Real] = [-0.05, 0.1],
        std: Iterable[Real] = [0.08, 0.32],
        per_channel: bool = False,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any
    ) -> NoReturn:
        """

        Parameters
        ----------
        mean: array_like, default [-0.05,0.1],
            range of mean value of the re-normalized signal, of shape (2,);
            or range of mean values for each lead of the re-normalized signal, of shape (lead, 2).
        std: array_like, default [0.08,0.32],
            range of standard deviation of the re-normalized signal, of shape (2,);
            or range of standard deviations for each lead of the re-normalized signal, of shape (lead, 2).
        per_channel: bool, default False,
            if True, re-normalization will be done per channel
        prob: float, default 0.5,
            Probability of applying the random re-normalization augmenter.
        inplace: bool, default True,
            Whether to apply the random re-normalization augmenter in-place.
        kwargs: keyword arguments
        """
        super().__init__()
        self.mean = np.array(mean)
        self.mean_mean = self.mean.mean(axis=-1, keepdims=True)
        self.mean_scale = (self.mean[..., -1] - self.mean_mean) * 0.3
        self.std = np.array(std)
        self.std_mean = self.std.mean(axis=-1, keepdims=True)
        self.std_scale = (self.std[..., -1] - self.std_mean) * 0.3
        self.per_channel = per_channel
        if not self.per_channel:
            assert self.mean.ndim == 1 and self.std.ndim == 1
        self.prob = prob
        self.inplace = inplace

    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """

        Parameters
        ----------
        sig: Tensor,
            the input ECG tensor, of shape (batch, lead, siglen)
        label: Tensor, optional,
            the input ECG label tensor,
            not used, but kept for compatibility with other augmenters
        extra_tensors: sequence of Tensors, optional,
            not used, but kept for consistency with other augmenters
        kwargs: keyword arguments,
            not used, but kept for consistency with other augmenters

        Returns
        -------
        sig: Tensor,
            the randomly re-normalized ECG tensor.
        label: Tensor,
            the label tensor of the augmented ECGs, unchanged
        extra_tensors: sequence of Tensors, optional,
            if set in the input arguments, unchanged
        """
        batch, lead, siglen = sig.shape
        if self.mean.ndim == 2:
            assert self.mean.shape[0] == lead
        if self.std.ndim == 2:
            assert self.std.shape[0] == lead
        if not self.inplace:
            sig = sig.clone()
        indices = self.get_indices(self.prob, pop_size=batch)
        if self.per_channel:
            mean = DEFAULTS.RNG.normal(
                self.mean_mean, self.mean_scale, size=(len(indices), lead, 1)
            )
            std = DEFAULTS.RNG.normal(
                self.std_mean, self.std_scale, size=(len(indices), lead, 1)
            )
        else:
            mean = DEFAULTS.RNG.normal(
                self.mean_mean, self.mean_scale, size=(len(indices), 1, 1)
            )
            std = DEFAULTS.RNG.normal(
                self.std_mean, self.std_scale, size=(len(indices), 1, 1)
            )
        sig[indices, ...] = normalize_t(
            sig[indices, ...],
            method="z-score",
            mean=mean,
            std=std,
            per_channel=self.per_channel,
            inplace=True,
        )
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        """ """
        return [
            "mean",
            "std",
            "per_channel",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
Standardized ECG diagnostic statements by AHA/ACC/HRS (i.e. the American Heart Association, the American College of Cardiology, and the Heart Rhythm Society)

References
----------
1. Mason, J. W., Hancock, E. W. & Gettes, L. S. Recommendations for the standardization and interpretation of the electrocardiogram.
Circulation 115, 1325–1332, https://doi.org/10.1161/CIRCULATIONAHA.106.180201 (2007).
"""

import io

import pandas as pd


__all__ = [
    "df_primary_statements",
    "df_secondary_statements",
    "df_modifiers",
    "df_comparison_statements",
    "df_convenience_statements",
    "df_secondary_primary_statement_pairing_rules",
    "df_general_modifier_primary_statement_pairing_rules",
]


df_primary_statements = pd.read_csv(
    io.StringIO(
        """
CategoryCode,Category,Code,Description
A,Overall interpretation,1,Normal ECG
,,2,Otherwise normal ECG
,,3,Abnormal ECG
,,4,Uninterpretable ECG
B,Technical conditions,10,Extremity electrode reversal
,,11,Misplaced precordial electrode(s)
,,12,Missing lead(s)
,,13,Right-sided precordial electrode(s)
,,14,Artifact
,,15,Poor-quality data
,,16,Posterior electrode(s)
C,Sinus node rhythms and arrhythmias,20,Sinus rhythm
,,21,Sinus tachycardia
,,22,Sinus bradycardia
,,23,Sinus arrhythmia
,,24,"Sinoatrial block, type I"
,,25,"Sinoatrial block, type II"
,,26,Sinus pause or arrest
,,27,Uncertain supraventricular rhythm
D,Supraventricular arrhythmias,30,Atrial premature complex(es)
,,31,"Atrial premature complexes, nonconducted"
,,32,Retrograde atrial activation
,,33,Wandering atrial pacemaker
,,34,Ectopic atrial rhythm
,,35,"Ectopic atrial rhythm, multifocal"
,,36,Junctional premature complex(es)
,,37,Junctional escape complex(es)
,,38,Junctional rhythm
,,39,Accelerated junctional rhythm
,,40,Supraventricular rhythm
,,41,Supraventricular complex(es)
,,42,"Bradycardia, nonsinus"
E,Supraventricular tachyarrhythmias,50,Atrial fibrillation
,,51,Atrial flutter
,,52,"Ectopic atrial tachycardia, unifocal"
,,53,"Ectopic atrial tachycardia, multifocal"
,,54,Junctional tachycardia
,,55,Supraventricular tachycardia
,,56,Narrow-QRS tachycardia
F,Ventricular arrhythmias,60,Ventricular premature complex(es)
,,61,Fusion complex(es)
,,62,Ventricular escape complex(es)
,,63,Idioventricular rhythm
,,64,Accelerated idioventricular rhythm
,,65,Fascicular rhythm
,,66,Parasystole
G,Ventricular tachyarrhythmias,70,Ventricular tachycardia
,,71,"Ventricular tachycardia, unsustained"
,,72,"Ventricular tachycardia, polymorphous"
,,73,"Ventricular tachycardia, torsades de pointes"
,,74,Ventricular fibrillation
,,75,Fascicular tachycardia
,,76,Wide-QRS tachycardia
H,Atrioventricular conduction,80,Short PR interval
,,81,AV conduction ratio N:D
,,82,Prolonged PR interval
,,83,"Second-degree AV block, Mobitz type I (Wenckebach)"
,,84,"Second-degree AV block, Mobitz type II"
,,85,2:1 AV block
,,86,"AV block, varying conduction"
,,87,"AV block, advanced (high-grade)"
,,88,"AV block, complete (third-degree)"
,,89,AV dissociation
I,Intraventricular and intra-atrial conduction,100,Aberrant conduction of supraventricular beat(s)
,,101,Left anterior fascicular block
,,102,Left posterior fascicular block
,,104,Left bundle-branch block
,,105,Incomplete right bundle-branch block
,,106,Right bundle-branch block
,,107,Intraventricular conduction delay
,,108,Ventricular preexcitation
,,109,Right atrial conduction abnormality
,,110,Left atrial conduction abnormality
,,111,Epsilon wave
J,Axis and voltage,120,Right-axis deviation
,,121,Left-axis deviation
,,122,Right superior axis
,,123,Indeterminate axis
,,124,Electrical alternans
,,125,Low voltage
,,128,Abnormal precordial R-wave progression
,,131,Abnormal P-wave axis
K,Chamber hypertrophy or enlargement,140,Left atrial enlargement
,,141,Right atrial enlargement
,,142,Left ventricular hypertrophy
,,143,Right ventricular hypertrophy
,,144,Biventricular hypertrophy
L,"ST segment, T wave, and U wave",145,ST deviation
,,146,ST deviation with T-wave change
,,147,T-wave abnormality
,,148,Prolonged QT interval
,,149,Short QT interval
,,150,Prominent U waves
,,151,Inverted U waves
,,152,TU fusion
,,153,ST-T change due to ventricular ?hypertrophy
,,154,Osborn wave
,,155,Early repolarization
M,Myocardial infarction,160,Anterior MI
,,161,Inferior MI
,,162,Posterior MI
,,163,Lateral MI
,,165,Anteroseptal MI
,,166,Extensive anterior MI
,,173,MI in presence of left bundle-branch ?block
,,174,Right ventricular MI
N,Pacemaker,180,Atrial-paced complex(es) or rhythm
,,181,Ventricular-paced complex(es) or rhythm
,,182,Ventricular pacing of non–right ventricular ?apical origin
,,183,Atrial-sensed ventricular-paced ?complex(es) or rhythm
,,184,AV dual-paced complex(es) or rhythm
,,185,"Failure to capture, atrial"
,,186,"Failure to capture, ventricular"
,,187,"Failure to inhibit, atrial"
,,188,"Failure to inhibit, ventricular"
,,189,"Failure to pace, atrial"
,,190,"Failure to pace, ventricular"
"""
    ),
    dtype=str,
)

df_primary_statements = df_primary_statements.fillna(method="ffill")


df_secondary_statements = pd.read_csv(
    io.StringIO(
        """
Group,Code,Description
Suggests,200,Acute pericarditis
,201,Acute pulmonary embolism
,202,Brugada abnormality
,203,Chronic pulmonary disease
,204,CNS disease
,205,Digitalis effect
,206,Digitalis toxicity
,207,Hypercalcemia
,208,Hyperkalemia
,209,Hypertrophic cardiomyopathy
,210,Hypocalcemia
,211,Hypokalemia or drug effect
,212,Hypothermia
,213,Ostium primum ASD
,214,Pericardial effusion
,215,Sinoatrial disorder
Consider,220,Acute ischemia
,221,AV nodal reentry
,222,AV reentry
,223,Genetic repolarization abnormality
,224,High precordial lead placement
,225,Hypothyroidism
,226,Ischemia
,227,Left ventricular aneurysm
,228,Normal variant
,229,Pulmonary disease
,230,Dextrocardia
,231,Dextroposition
"""
    ),
    dtype=str,
)

df_secondary_statements = df_secondary_statements.fillna(method="ffill")


df_modifiers = pd.read_csv(
    io.StringIO(
        """
Category,Code,Description,
General,301,Borderline,
,303,Increased,
,304,Intermittent,
,305,Marked,
,306,Moderate,
,307,Multiple,
,308,Occasional,
,309,One,
,310,Frequent,
,312,Possible,
,313,Postoperative,
,314,Predominant,
,315,Probable,
,316,Prominent,
,317,(Specified) Lead(s),
,318,(Specified) Electrode(s),
,321,Nonspecific,
General: conjunctions,302,Consider,
,310,Or,
,320,And,
,319,With,
,322,Versus,
Myocardial infarction,330,Acute,
,331,Recent,
,332,Old,
,333,Of indeterminate age,
,334,Evolving,
Arrhythmias and tachyarrhythmias,340,Couplets,
,341,In a bigeminal pattern,
,342,In a trigeminal pattern,
,343,Monomorphic,
,344,Multifocal,
,345,Unifocal,
,346,With a rapid ventricular response,
,347,With a slow ventricular response,
,348,With capture beat(s),
,349,With aberrancy,
,350,Polymorphic,
Repolarization abnormalities,360,≥0.1 mV,
,361,≥0.2 mV,
,362,Depression,
,363,Elevation,
,364,Maximally toward lead,
,365,Maximally away from lead,
,366,Low amplitude,
,367,Inversion,
,369,Postpacing (anamnestic),
"""
    ),
    dtype=str,
)

df_modifiers = df_modifiers.fillna(method="ffill")


df_comparison_statements = pd.read_csv(
    io.StringIO(
        """
Code,Statement,Criteria
400,No significant change,"Intervals (PR, QRS, QTc) remain normal or within 10% of a previously abnormal value"
,,No new or deleted diagnoses with the exception of normal variant diagnoses
401,Significant change in rhythm,New or deleted rhythm diagnosis
,,HR change >20 bpm and <50 or >100 bpm
,,New or deleted pacemaker diagnosis
402,New or worsened ischemia or infarction,"Added infarction, ST-ischemia, or T-wave-ischemia diagnosis, or worsened ST deviation or T-wave abnormality"
403,New conduction abnormality,Added AV or IV conduction diagnosis
404,Significant repolarization change,New or deleted QT diagnosis
,,New or deleted U-wave diagnosis
,,New or deleted nonischemic ST or T-wave diagnosis
,,Change in QTc >60 ms
405,Change in clinical status,"New or deleted diagnosis from Axis and Voltage, Chamber Hypertrophy, or Enlargement primary statement categories or “Suggests…” secondary statement category"
406,Change in interpretation without significant change in waveform,"Used when a primary or secondary statement is added or removed despite no real change in the tracing; ie, an interpretive disagreement exists between the readers of the first and second ECGs"
"""
    ),
    dtype=str,
)

df_comparison_statements = df_comparison_statements.fillna(method="ffill")


df_convenience_statements = pd.read_csv(
    io.StringIO(
        """
Code,Statement
500,Nonspecific ST-T abnormality
501,ST elevation
502,ST depression
503,LVH with ST-T changes
"""
    ),
    dtype=str,
)


df_secondary_primary_statement_pairing_rules = pd.read_csv(
    io.StringIO(
        """
Secondary Code,May Accompany These Primary Codes
200,145-147
201,"21, 105, 109, 120, 131, 141, 145-147"
202,"105, 106, 145-146"
203,"109, 120, 125, 128, 131, 141, 143"
204,147
205,145-147
206,145-147
207,149
208,147
209,142
210,148
211,"147-148, 150"
212,"14, 154"
213,"82, 105-106, 121"
214,124
215,"42, 131, 145-147"
220,"145-147, 151"
221,"55, 56"
222,"55, 56"
223,"148, 149"
224,128
225,"22, 24-26, 37, 38"
226,145-147
227,145-147
228,"80, 105, 128, 155"
229,"109, 120, 122-123, 125, 128, 131, 141, 143"
230,"128, 131"
231,128
"""
    ),
    dtype=str,
)


df_general_modifier_primary_statement_pairing_rules = pd.read_csv(
    io.StringIO(
        """
General Modifier Code,May (May Not) Accompany These Primary Codes or May Be Between Codes in These Categories or Groups of Categories,May/May Not,Location
301,"1-20, 24-76, 81, 83-106, 108, 122-124",May not,b
302,"1-3, 12-16, 80-82, 111-130, 145-152",May not,"b, i"
303,"30, 31, 36, 37, 41, 60, 62, 63, 82, 107, 109, 110",May,"a, b"
304,"21-26, 30-76, 80, 82-108, 124, 180-190",May,b
305,"1-20, 27-76, 81, 85-106, 111, 122, 123, 148-150, 160-190",May not,b
306,"1-20, 27-76, 81, 85-106, 111, 122, 123, 148-150, 160-190",May not,b
307,"26, 30, 31, 36, 37, 41, 60-62, 185-190",May,b
308,"26, 30, 31, 36, 37, 41, 60-62, 185-190",May,b
309,"26, 30, 31, 36, 37, 41, 60-62, 185-190",May,b
310,"C, D, E, F, G, N, H, I, J, K, L, M",May,i
312,"1-3, 15, 80-82, 120-122, 128",May not,b
313,145-147,May,b
314,"20-23, 33-35, 38-56, 63-76, 83-89, 180-184",May,b
315,"1-3, 15, 80-82, 120-122, 128",May not,b
316,"1-20, 27-76, 81, 85-106, 111, 122, 123, 148-150, 160-190",May not,b
317,"C, D, E, F, G, N, H, I, J, K, L, M",May,i
318,"C, D, E, F, G, N, H, I, J, K, L, M",May,i
319,"C, D, E, F, G, N, 100, J, K, L, M",May,i
321,"40, 55, 56, 145-147",May,b
"""
    ),
    dtype=str,
)

df_general_modifier_primary_statement_pairing_rules[
    "Location"
] = df_general_modifier_primary_statement_pairing_rules["Location"].apply(
    lambda location: location.replace("b", "before")
    .replace("a", "after")
    .replace("i", "between")
)



"""
"""

import time
from typing import NoReturn

import torch

try:
    import torch_ecg  # noqa: F401
except ModuleNotFoundError:
    import sys
    from pathlib import Path

    sys.path.insert(0, str(Path(__file__).absolute().parents[1]))

from torch_ecg.cfg import CFG
from torch_ecg.model_configs import (  # noqa: F401
    ATI_CNN_CONFIG,
    ECG_CRNN_CONFIG,
    ECG_SEQ_LAB_NET_CONFIG,
    ECG_SUBTRACT_UNET_CONFIG,
    ECG_UNET_VANILLA_CONFIG,
    ECG_YOLO_CONFIG,
    RR_AF_CRF_CONFIG,
    RR_AF_VANILLA_CONFIG,
    RR_LSTM_CONFIG,
    attention,
    cpsc_2018,
    cpsc_2018_leadwise,
    densenet_leadwise,
    densenet_vanilla,
    global_context,
    linear,
    lstm,
    mobilenet_v1_vanilla,
    multi_scopic,
    multi_scopic_leadwise,
    non_local,
    resnet_block_stanford,
    resnet_cpsc2018,
    resnet_cpsc2018_leadwise,
    resnet_nature_comm,
    resnet_nature_comm_bottle_neck,
    resnet_nature_comm_bottle_neck_gc,
    resnet_nature_comm_bottle_neck_nl,
    resnet_nature_comm_bottle_neck_se,
    resnet_stanford,
    resnet_vanilla_18,
    resnet_vanilla_34,
    resnet_vanilla_50,
    resnet_vanilla_101,
    resnet_vanilla_152,
    resnet_vanilla_wide_50_2,
    resnet_vanilla_wide_101_2,
    resnetN,
    resnetNB,
    resnetNBS,
    resnetNS,
    resnext_vanilla_50_32x4d,
    resnext_vanilla_101_32x8d,
    squeeze_excitation,
    tresnetF,
    tresnetL,
    tresnetM,
    tresnetN,
    tresnetP,
    tresnetS,
    vgg16,
    vgg16_leadwise,
    xception_leadwise,
    xception_vanilla,
)
from torch_ecg.models.cnn.darknet import DarkNet  # noqa: F401
from torch_ecg.models.cnn.densenet import DenseNet  # noqa: F401
from torch_ecg.models.cnn.efficientnet import EfficientNet, EfficientNetV2  # noqa: F401
from torch_ecg.models.cnn.ho_resnet import (  # noqa: F401
    MidPointResNet,
    RK4ResNet,
    RK8ResNet,
)
from torch_ecg.models.cnn.mobilenet import (  # noqa: F401
    MobileNetV1,
    MobileNetV2,
    MobileNetV3,
)
from torch_ecg.models.cnn.multi_scopic import MultiScopicCNN  # noqa: F401
from torch_ecg.models.cnn.resnet import ResNet  # noqa: F401
from torch_ecg.models.cnn.vgg import VGG16  # noqa: F401
from torch_ecg.models.cnn.xception import Xception  # noqa: F401
from torch_ecg.models.ecg_crnn import ECG_CRNN  # noqa: F401
from torch_ecg.models.ecg_seq_lab_net import ECG_SEQ_LAB_NET  # noqa: F401
from torch_ecg.models.rr_lstm import RR_LSTM  # noqa: F401
from torch_ecg.models.unets.ecg_subtract_unet import ECG_SUBTRACT_UNET  # noqa: F401
from torch_ecg.models.unets.ecg_unet import ECG_UNET  # noqa: F401

_DEVICE = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

_IN_CHANNELS = 12
_BATCH_SIZE = 2
_SIG_LEN = 4000
_RR_LEN = 100
_TEST_EXAMPLE = torch.rand((_BATCH_SIZE, _IN_CHANNELS, _SIG_LEN)).to(_DEVICE)
_TEST_RR_EXAMPLE = torch.rand(
    (
        _RR_LEN,
        _BATCH_SIZE,
        1,
    )
).to(_DEVICE)
_TEST_CLF_CLASSES = [
    "nsr",
    "af",
    "pvc",
]
_TEST_DELI_CLASSES = [
    "qrs",
    "p",
    "t",
]


@torch.no_grad()
def test_cnn() -> NoReturn:
    """ """
    start = time.time()
    print("\n" + " Test CNN configs ".center(80, "#") + "\n")
    # test vgg configs
    print("\n" + " Test VGG configs ".center(50, "-") + "\n")
    VGG16.__DEBUG__ = False
    for cfg in [
        "vgg16",
        "vgg16_leadwise",
    ]:
        _test_cnn("VGG16", cfg)

    # test resnet configs
    print("\n" + " Test ResNet configs ".center(50, "-") + "\n")
    ResNet.__DEBUG__ = False
    for cfg in [
        "resnet_vanilla_18",
        "resnet_vanilla_34",
        "resnet_vanilla_50",
        "resnet_vanilla_101",
        "resnet_vanilla_152",
        "resnet_vanilla_wide_50_2",
        "resnet_vanilla_wide_101_2",
        # custom resnet
        "resnetN",
        "resnetNB",
        "resnetNBS",
        "resnetNS",
        "tresnetN",
        "tresnetP",
        "tresnetF",
        "tresnetS",
        "tresnetM",
        "tresnetL",
        "resnet_nature_comm",
        "resnet_nature_comm_bottle_neck",
        "resnet_nature_comm_bottle_neck_se",
        "resnet_nature_comm_bottle_neck_nl",
        "resnet_nature_comm_bottle_neck_gc",
        # "resnet_leadwise",
        # TODO: fix bugs in the following
        # "resnext_vanilla_50_32x4d", "resnext_vanilla_101_32x8d",
        # stanford resnet
        # "resnet_stanford",
    ]:
        _test_cnn("ResNet", cfg)

    # test multi_scopic
    print("\n" + " Test MultiScopic configs ".center(50, "-") + "\n")
    MultiScopicCNN.__DEBUG__ = False
    for cfg in [
        "multi_scopic",
        "multi_scopic_leadwise",
    ]:
        _test_cnn("MultiScopicCNN", cfg)

    # test densenet
    print("\n" + " Test DenseNet configs ".center(50, "-") + "\n")
    DenseNet.__DEBUG__ = False
    for cfg in [
        "densenet_vanilla",
        "densenet_leadwise",
    ]:
        _test_cnn("DenseNet", cfg)

    # test xception
    print("\n" + " Test Xception configs ".center(50, "-") + "\n")
    Xception.__DEBUG__ = False
    for cfg in [
        "xception_vanilla",
        "xception_leadwise",
    ]:
        _test_cnn("Xception", cfg)

    # test mobilenet
    print("\n" + " Test MobileNet configs ".center(50, "-") + "\n")
    MobileNetV1.__DEBUG__ = False
    MobileNetV2.__DEBUG__ = False
    MobileNetV3.__DEBUG__ = False
    for cfg in [
        "mobilenet_v1_vanilla",
    ]:
        _test_cnn("MobileNetV1", cfg)

    print(f"total time cost: {time.time()-start:.2f} seconds")
    print("\n" + " Finish testing CNN configs ".center(80, "#") + "\n")


@torch.no_grad()
def _test_cnn(model_name: str, cfg: CFG) -> NoReturn:
    """ """
    try:
        test_model = eval(
            f"{model_name}(in_channels=_IN_CHANNELS, **{cfg}).to(_DEVICE)"
        )
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"{cfg} output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print(f"{cfg} raises errors\n")
        raise e


@torch.no_grad()
def test_tasks() -> NoReturn:
    """ """
    start = time.time()
    print("\n" + " Test downstream task configs ".center(80, "#") + "\n")
    # test crnn configs
    print("\n" + " Test ECG_CRNN configs ".center(50, "-") + "\n")
    ECG_CRNN.__DEBUG__ = False
    try:
        test_model = ECG_CRNN(
            classes=_TEST_CLF_CLASSES, n_leads=_IN_CHANNELS, config=ECG_CRNN_CONFIG
        ).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_CRNN output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_CRNN_CONFIG raises errors\n")
        raise e

    # test seq_lab configs
    print("\n" + " Test ECG_SEQ_LAB_NET configs ".center(50, "-") + "\n")
    ECG_SEQ_LAB_NET.__DEBUG__ = False
    try:
        test_model = ECG_SEQ_LAB_NET(
            classes=_TEST_DELI_CLASSES,
            n_leads=_IN_CHANNELS,
            config=ECG_SEQ_LAB_NET_CONFIG,
        ).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_SEQ_LAB_NET output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_SEQ_LAB_NET raises errors\n")
        raise e

    # test unet configs
    print("\n" + " Test ECG_UNET configs ".center(50, "-") + "\n")
    ECG_UNET.__DEBUG__ = False
    try:
        test_model = ECG_UNET(
            classes=_TEST_DELI_CLASSES,
            n_leads=_IN_CHANNELS,
            config=ECG_UNET_VANILLA_CONFIG,
        ).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_UNET output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_UNET_VANILLA_CONFIG raises errors\n")
        raise e

    # test subtract_unet configs
    print("\n" + " Test ECG_SUBTRACT_UNET configs ".center(50, "-") + "\n")
    ECG_SUBTRACT_UNET.__DEBUG__ = False
    try:
        test_model = ECG_SUBTRACT_UNET(
            classes=_TEST_DELI_CLASSES,
            n_leads=_IN_CHANNELS,
            config=ECG_SUBTRACT_UNET_CONFIG,
        ).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_SUBTRACT_UNET output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_SUBTRACT_UNET_CONFIG raises errors\n")
        raise e

    # test rr_lstm configs
    print("\n" + " Test RR_LSTM configs ".center(50, "-") + "\n")
    RR_LSTM.__DEBUG__ = False
    for cfg in [
        "RR_AF_CRF_CONFIG",
        "RR_AF_VANILLA_CONFIG",
        "RR_LSTM_CONFIG",
    ]:
        try:
            test_model = eval(
                f"RR_LSTM(classes=_TEST_CLF_CLASSES, config={cfg}).to(_DEVICE)"
            )
            test_model.eval()
            test_output = test_model(_TEST_RR_EXAMPLE)
            print(f"{cfg} output shape = {test_output.shape}")
            del test_model
            del test_output
        except Exception as e:
            print(f"{cfg} raises errors\n")
            raise e

    print(f"total time cost: {time.time()-start:.2f} seconds")
    print("\n" + " Finish testing downstream task configs ".center(80, "#") + "\n")


def run_test():
    """ """
    # test CNNs
    test_cnn()
    # test downstream tasks
    test_tasks()


if __name__ == "__main__":
    run_test()



"""
neural architecture search

"""

from typing import NoReturn, Sequence

import torch
import torch.nn as nn
from torch.nn.parallel import DataParallel as DP
from torch.nn.parallel import DistributedDataParallel as DDP  # noqa: F401
from torch.utils.data import DataLoader  # noqa: F401
from torch.utils.data.dataset import Dataset

from ..cfg import CFG
from .trainer import BaseTrainer


__all__ = [
    "NAS",
]


class NAS:
    """ """

    __name__ = "NAS"

    def __init__(
        self,
        trainer_cls: BaseTrainer,
        model_cls: nn.Module,
        dataset_cls: Dataset,
        train_config: dict,
        model_configs: Sequence[dict],
        lazy: bool = False,
    ) -> NoReturn:
        """

        Parameters
        ----------
        trainer_cls: BaseTrainer,
            trainer class
        model_cls: nn.Module,
            model class
        dataset_cls: Dataset,
            dataset class
        train_config: dict,
            train configurations
        model_configs: sequence of dict,
            model configurations, each with a different network architecture
        lazy: bool, default False,
            whether to load the dataset in the trainer at initialization

        """
        self.trainer_cls = trainer_cls
        self.model_cls = model_cls
        self.dataset_cls = dataset_cls
        self.train_config = CFG(train_config)
        self.model_configs = model_configs
        self.lazy = lazy
        if not lazy:
            self.ds_train = self.dataset_cls(
                self.train_config, training=True, lazy=False
            )
            self.ds_val = self.dataset_cls(
                self.train_config, training=False, lazy=False
            )
        else:
            self.ds_train = None
            self.ds_val = None

    def search(self) -> NoReturn:
        """ """
        if self.ds_train is None or self.ds_val is None:
            raise ValueError("training dataset or validation dataset is not set")
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        for model_config in self.model_configs:
            model = self.model_cls(
                classes=self.train_config.classes,
                n_leads=self.train_config.n_leads,
                config=model_config,
            )
            if torch.cuda.device_count() > 1:
                model = DP(model)
                # model = DDP(model)
            model.to(device=device)
            model.train()
            trainer = self.trainer_cls(
                model=model,
                dataset_cls=self.dataset_cls,
                train_config=self.train_config,
                model_config=model_config,
                device=device,
                lazy=True,
            )
            trainer._setup_dataloaders(self.ds_train, self.ds_val)
            trainer.train()

            del model
            del trainer
            torch.cuda.empty_cache()

    def _setup_dataset(self, ds_train: Dataset, ds_val: Dataset) -> NoReturn:
        """

        Parameters
        ----------
        ds_train: Dataset,
            training dataset
        ds_val: Dataset,
            validation dataset

        """
        self.ds_train = ds_train
        self.ds_val = ds_val



# -*- coding: utf-8 -*-
"""
"""

import math
from copy import deepcopy
from numbers import Real
from pathlib import Path
from typing import Any, Dict, List, NoReturn, Optional, Sequence, Union

import numpy as np
import wfdb
from scipy.signal import resample_poly

from ...cfg import CFG, DEFAULTS
from ...utils.misc import add_docstring, get_record_list_recursive
from ...utils.utils_interval import generalized_intervals_intersection
from ..base import DEFAULT_FIG_SIZE_PER_SEC, PhysioNetDataBase, DataBaseInfo


__all__ = [
    "AFDB",
]


_AFDB_INFO = DataBaseInfo(
    title="""
    MIT-BIH Atrial Fibrillation Database
    """,
    about="""
    1. contains 25 long-term (each 10 hours) ECG recordings of human subjects with atrial fibrillation (mostly paroxysmal)
    2. 23 records out of 25 include the two ECG signals, the left 2 records 00735 and 03665 are represented only by the rhythm (.atr) and unaudited beat (.qrs) annotation files
    3. signals are sampled at 250 samples per second with 12-bit resolution over a range of ±10 millivolts, with a typical recording bandwidth of approximately 0.1 Hz to 40 Hz
    4. 4 classes of rhythms are annotated:
        - AFIB:  atrial fibrillation
        - AFL:   atrial flutter
        - J:     AV junctional rhythm
        - N:     all other rhythms
    5. rhythm annotations almost all start with "(N", except for 4 which start with '(AFIB', which are all within 1 second (250 samples)
    """,
    note="""
    1. beat annotation files (.qrs files) were prepared using an automated detector and have NOT been corrected manually
    2. for some records, manually corrected beat annotation files (.qrsc files) are available
    3. one should never use wfdb.rdann with arguments `sampfrom`, since one has to know the `aux_note` (with values in ["(N", "(J", "(AFL", "(AFIB"]) before the index at `sampfrom`
    """,
    usage=[
        "Atrial fibrillation (AF) detection",
    ],
    references=[
        "https://physionet.org/content/afdb/1.0.0/",
        "Moody GB, Mark RG. A new method for detecting atrial fibrillation using R-R intervals. Computers in Cardiology. 10:227-230 (1983).",
    ],
    doi=[
        "10.13026/C2MW2D",
    ],
)


@add_docstring(_AFDB_INFO.format_database_docstring())
class AFDB(PhysioNetDataBase):
    """ """

    __name__ = "AFDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[str] = None,
        verbose: int = 2,
        **kwargs: Any,
    ) -> NoReturn:
        """

        Parameters
        ----------
        db_dir: str or Path, optional,
            storage path of the database
            if not specified, data will be fetched from Physionet
        working_dir: str, optional,
            working directory, to store intermediate files and log file
        verbose: int, default 2,
            log verbosity
        kwargs: auxilliary key word arguments

        """
        super().__init__(
            db_name="afdb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 250
        self.data_ext = "dat"
        self.ann_ext = "atr"
        self.auto_beat_ann_ext = "qrs"
        self.manual_beat_ann_ext = "qrsc"

        self.all_leads = [
            "ECG1",
            "ECG2",
        ]

        self._ls_rec()
        self.special_records = ["00735", "03665"]
        self.qrsc_records = get_record_list_recursive(
            self.db_dir, self.manual_beat_ann_ext
        )

        self.class_map = CFG(AFIB=1, AFL=2, J=3, N=0)  # an extra isoelectric
        self.palette = kwargs.get("palette", None)
        if self.palette is None:
            self.palette = CFG(
                AFIB="blue",
                AFL="red",
                J="yellow",
                # N="green",
                qrs="green",
            )

    def load_data(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, List[str]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: str = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """
        load physical (converted from digital) ecg data,
        which is more understandable for humans

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        leads: str or list of str, optional,
            the leads to load
        sampfrom: int, optional,
            start index of the data to be loaded
        sampto: int, optional,
            end index of the data to be loaded
        data_format: str, default "channel_first",
            format of the ecg data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first")
        units: str, default "mV",
            units of the output signal, can also be "μV", with an alias of "uV"
        fs: real number, optional,
            if not None, the loaded data will be resampled to this frequency

        Returns
        -------
        data: ndarray,
            the ecg data

        """
        fp = str(self.get_absolute_path(rec))
        if not leads:
            _leads = self.all_leads
        elif isinstance(leads, str):
            _leads = [leads]
        else:
            _leads = leads
        assert set(_leads).issubset(self.all_leads)
        # p_signal in the format of "lead_last", and in units "mV"
        data = wfdb.rdrecord(
            fp,
            sampfrom=sampfrom or 0,
            sampto=sampto,
            physical=True,
            channel_names=_leads,
        ).p_signal.astype(DEFAULTS.np_dtype)
        if units.lower() in ["μv", "uv"]:
            data = 1000 * data
        if fs is not None and fs != self.fs:
            data = resample_poly(data, fs, self.fs, axis=0).astype(DEFAULTS.np_dtype)
        if data_format.lower() in ["channel_first", "lead_first"]:
            data = data.T
        return data

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        ann_format: str = "interval",
        keep_original: bool = False,
    ) -> Union[Dict[str, list], np.ndarray]:
        """
        load annotations (header) stored in the .hea files

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        ann_format: str, default "interval", case insensitive,
            format of returned annotation, can also be "mask"
        keep_original: bool, default False,
            if True, in the "interval" `ann_format`,
            intervals (in the form [a,b]) will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified

        Returns
        -------
        ann, dict or ndarray,
            the annotations in the format of intervals, or in the format of mask

        """
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=self.ann_ext)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        ann = CFG({k: [] for k in self.class_map.keys()})
        critical_points = wfdb_ann.sample.tolist() + [sig_len]
        aux_note = wfdb_ann.aux_note
        if aux_note[0] == "(N":
            # ref. the doc string of the class
            critical_points[0] = 0
        else:
            critical_points.insert(0, 0)
            aux_note.insert(0, "(N")

        for idx, rhythm in enumerate(aux_note):
            ann[rhythm.replace("(", "")].append(
                [critical_points[idx], critical_points[idx + 1]]
            )
        ann = CFG(
            {
                k: generalized_intervals_intersection(l_itv, [[sf, st]])
                for k, l_itv in ann.items()
            }
        )

        if ann_format.lower() == "mask":
            tmp = deepcopy(ann)
            ann = np.full(shape=(st - sf,), fill_value=self.class_map.N, dtype=int)
            for rhythm, l_itv in tmp.items():
                for itv in l_itv:
                    ann[itv[0] - sf : itv[1] - sf] = self.class_map[rhythm]
        elif not keep_original:
            for k, l_itv in ann.items():
                ann[k] = [[itv[0] - sf, itv[1] - sf] for itv in l_itv]

        return ann

    def load_beat_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        use_manual: bool = True,
        keep_original: bool = False,
    ) -> np.ndarray:
        """
        load beat annotations stored in corresponding annotation files

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        use_manual: bool, default True,
            use manually annotated beat annotations (qrs),
            instead of those generated by algorithms
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified

        Returns
        -------
        ann, ndarray,
            locations (indices) of the qrs complexes

        """
        if isinstance(rec, int):
            rec = self[rec]
        fp = str(self.get_absolute_path(rec))
        if use_manual and rec in self.qrsc_records:
            ext = self.manual_beat_ann_ext
        else:
            ext = self.auto_beat_ann_ext
        ann = wfdb.rdann(
            fp,
            extension=ext,
            sampfrom=sampfrom or 0,
            sampto=sampto,
        )
        ann = ann.sample
        if not keep_original and sampfrom is not None:
            ann -= sampfrom
        return ann

    @add_docstring(load_beat_ann.__doc__)
    def load_rpeak_indices(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        use_manual: bool = True,
        keep_original: bool = False,
    ) -> np.ndarray:
        """
        alias of `self.load_beat_ann`
        """
        return self.load_beat_ann(rec, sampfrom, sampto, use_manual, keep_original)

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ann: Optional[Dict[str, np.ndarray]] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[str, List[str]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        same_range: bool = False,
        **kwargs: Any,
    ) -> NoReturn:
        """

        plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        data: ndarray, optional,
            (2-lead) ecg signal to plot,
            should be of the format "channel_first", and compatible with `leads`
            if given, data of `rec` will not be used,
            this is useful when plotting filtered data
        ann: dict, optional,
            annotations for `data`, covering those from annotation files,
            in the form of {"AFIB":l_itv, "AFL":l_itv, "J":l_itv, "N":l_itv},
            where `l_itv` in the form of [[a,b], ...],
            ignored if `data` is None
        rpeak_inds: array_like, optional,
            indices of R peaks, covering those from annotation files,
            if `data` is None, then indices should be the absolute indices in the record,
        ticks_granularity: int, default 0,
            the granularity to plot axis ticks, the higher the more,
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads: str or list of str, optional,
            the leads to plot
        sampfrom: int, optional,
            start index of the data to plot
        sampto: int, optional,
            end index of the data to plot
        same_range: bool, default False,
            if True, forces all leads to have the same y range
        kwargs: dict,
            keyword arguments for `matplotlib.pyplot.plot`, etc.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000
        if leads is None or leads == "all":
            _leads = self.all_leads
        elif isinstance(leads, str):
            _leads = [leads]
        else:
            _leads = leads
        assert all([ld in self.all_leads for ld in _leads])

        lead_indices = [self.all_leads.index(ld) for ld in _leads]
        if data is None:
            _data = self.load_data(
                rec,
                leads=_leads,
                sampfrom=sampfrom,
                sampto=sampto,
                data_format="channel_first",
                units="μV",
            )
        else:
            units = self._auto_infer_units(data)
            print(f"input data is auto detected to have units in {units}")
            if units.lower() == "mv":
                _data = 1000 * data
            else:
                _data = data
            _leads = [f"ECG_{idx}" for idx in range(_data.shape[0])]
        if ann is None and data is None:
            _ann = self.load_ann(
                rec,
                sampfrom=sampfrom,
                sampto=sampto,
                ann_format="interval",
                keep_original=False,
            )
        else:
            _ann = ann or CFG({k: [] for k in self.class_map.keys()})
        # indices to time
        _ann = {
            k: [[itv[0] / self.fs, itv[1] / self.fs] for itv in l_itv]
            for k, l_itv in _ann.items()
        }
        if rpeak_inds is None and data is None:
            _rpeak = self.load_rpeak_indices(
                rec,
                sampfrom=sampfrom,
                sampto=sampto,
                use_manual=True,
                keep_original=False,
            )
            _rpeak = _rpeak / self.fs  # indices to time
        else:
            _rpeak = np.array(rpeak_inds or []) / self.fs  # indices to time

        ann_plot_alpha = 0.2
        rpeaks_plot_alpha = 0.8

        nb_leads = len(_leads)

        line_len = self.fs * 25  # 25 seconds
        nb_lines = math.ceil(_data.shape[1] / line_len)

        for seg_idx in range(nb_lines):
            seg_data = _data[..., seg_idx * line_len : (seg_idx + 1) * line_len]
            secs = (np.arange(seg_data.shape[1]) + seg_idx * line_len) / self.fs
            seg_ann = {
                k: generalized_intervals_intersection(l_itv, [[secs[0], secs[-1]]])
                for k, l_itv in _ann.items()
            }
            seg_rpeaks = _rpeak[np.where((_rpeak >= secs[0]) & (_rpeak < secs[-1]))[0]]
            fig_sz_w = int(
                round(DEFAULT_FIG_SIZE_PER_SEC * seg_data.shape[1] / self.fs)
            )
            if same_range:
                y_ranges = (
                    np.ones((seg_data.shape[0],)) * np.max(np.abs(seg_data)) + 100
                )
            else:
                y_ranges = np.max(np.abs(seg_data), axis=1) + 100
            fig_sz_h = 6 * y_ranges / 1500
            fig, axes = plt.subplots(
                nb_leads, 1, sharex=True, figsize=(fig_sz_w, np.sum(fig_sz_h))
            )
            if nb_leads == 1:
                axes = [axes]
            for idx in range(nb_leads):
                axes[idx].plot(
                    secs, seg_data[idx], color="black", label=f"lead - {_leads[idx]}"
                )
                axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
                # NOTE that `Locator` has default `MAXTICKS` equal to 1000
                if ticks_granularity >= 1:
                    axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                    axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                    axes[idx].grid(
                        which="major", linestyle="-", linewidth="0.5", color="red"
                    )
                if ticks_granularity >= 2:
                    axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                    axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                    axes[idx].grid(
                        which="minor", linestyle=":", linewidth="0.5", color="black"
                    )
                for k, l_itv in seg_ann.items():
                    if k == "N":
                        continue
                    for itv in l_itv:
                        axes[idx].axvspan(
                            itv[0],
                            itv[1],
                            color=self.palette[k],
                            alpha=ann_plot_alpha,
                            label=k,
                        )
                for ri in seg_rpeaks:
                    axes[idx].axvspan(
                        ri - 0.01,
                        ri + 0.01,
                        color=self.palette["qrs"],
                        alpha=rpeaks_plot_alpha,
                    )
                    axes[idx].axvspan(
                        ri - 0.075,
                        ri + 0.075,
                        color=self.palette["qrs"],
                        alpha=ann_plot_alpha,
                    )
                axes[idx].legend(loc="upper left")
                axes[idx].set_xlim(secs[0], secs[-1])
                axes[idx].set_ylim(-y_ranges[idx], y_ranges[idx])
                axes[idx].set_xlabel("Time [s]")
                axes[idx].set_ylabel("Voltage [μV]")
            plt.subplots_adjust(hspace=0.2)
            plt.show()



# -*- coding: utf-8 -*-
"""
knowledge about ECG arrhythmia

Standard 12Leads ECG
--------------------
    Inferior leads: II, III, aVF
    Lateral leads: I, aVL, V5-6
    Septal leads: aVR, V1
    Anterior leads: V2-4
    -----------------------------------
    Chest (precordial) leads: V1-6
    Limb leads: I, II, III, aVR, aVL, aVF

ECG rhythm (https://litfl.com/ecg-rhythm-evaluation/)
-----------------------------------------------------
1. On a 12 lead ECG, ECG rhythm is usually a 10 second recording from Lead II
2. 7 steps to analyze:
    2.1. rate (brady < 60 bpm; 60 bpm <= normal <= 100 bpm; tachy > 100 bpm)
    2.2. pattern (temporal) of QRS complex (regular or irregular; if irregular, regularly irregular or irregularly irregular)
    2.3. morphology (spatial) of QRS complex (narrow <= 120 ms; wide > 120 ms)
    2.4. P waves (absent or present; morphology; PR interval)
    2.5. relation of P waves and QRS complexes (atrial rate and ventricular rate, AV association or AV disassociation)
    2.6. onset and termination (abrupt or gradual)
    (2.7. response to vagal manoeuvres)

ECG waves
---------
https://ecgwaves.com/topic/ecg-normal-p-wave-qrs-complex-st-segment-t-wave-j-point/

References
----------
[1] https://litfl.com/
[2] https://ecgwaves.com/
[3] https://ecglibrary.com/ecghome.php
[4] https://courses.cs.washington.edu/courses/cse466/13au/pdfs/lectures/ECG%20filtering.pdf

NOTE that wikipedia is NOT listed in the References

"""
from io import StringIO

import pandas as pd

from ..cfg import CFG

__all__ = [
    # named lead sets
    "Standard12Leads",
    "ChestLeads",
    "PrecordialLeads",
    "LimbLeads",
    "InferiorLeads",
    "LateralLeads",
    "SeptalLeads",
    "AnteriorLeads",
    # ECG abnormalities (arrhythmias)
    "AF",
    "AFL",  # atrial
    "IAVB",
    "LBBB",
    "CLBBB",
    "RBBB",
    "CRBBB",
    "IRBBB",
    "LAnFB",
    "NSIVCB",
    "BBB",  # conduction block
    "PAC",
    "PJC",
    "PVC",
    "SPB",  # premature: qrs, morphology
    "LPR",
    "LQT",
    "QAb",
    "TAb",
    "TInv",  # wave morphology
    "LAD",
    "RAD",  # axis
    "Brady",
    "LQRSV",
    "PRWP",  # qrs (RR interval, amplitude)
    "SA",
    "SB",
    "NSR",
    "STach",  # sinus
    "PR",  # pacer
    "STD",
    "STE",  # ST segments
]


Standard12Leads = [
    "I",
    "II",
    "III",
    "aVR",
    "aVL",
    "aVF",
    "V1",
    "V2",
    "V3",
    "V4",
    "V5",
    "V6",
]
ChestLeads = [f"V{n}" for n in range(1, 7)]
PrecordialLeads = ChestLeads
LimbLeads = [
    "I",
    "II",
    "III",
    "aVR",
    "aVL",
    "aVF",
]
InferiorLeads = [
    "II",
    "III",
    "aVF",
]
LateralLeads = [
    "I",
    "aVL",
    "V5",
    "V6",
]
SeptalLeads = [
    "aVR",
    "V1",
]
AnteriorLeads = [
    "V2",
    "V3",
    "V4",
]


_snomedbrowser_url = "https://snomedbrowser.com/Codes/Details/"


AF = CFG(
    {  # rr, morphology
        "fullname": "atrial fibrillation",
        "url": [
            "https://litfl.com/atrial-fibrillation-ecg-library/",
            "https://en.wikipedia.org/wiki/Atrial_fibrillation#Screening",
        ],
        "knowledge": [
            "irregularly irregular rhythm",
            "no P waves",
            "absence of an isoelectric baseline",
            "variable ventricular rate",
            "QRS complexes usually < 120 ms unless pre-existing bundle branch block, accessory pathway, or rate related aberrant conduction",
            "fibrillatory waves (f-wave) may be present and can be either fine (amplitude < 0.5mm) or coarse (amplitude > 0.5mm)",
            "fibrillatory waves (f-wave) may mimic P waves leading to misdiagnosis",
        ],
    }
)

AFL = CFG(
    {  # rr, morphology
        "fullname": "atrial flutter",
        "url": [
            "https://litfl.com/atrial-flutter-ecg-library/",
            "https://en.wikipedia.org/wiki/Atrial_flutter",
        ],
        "knowledge": [
            "a type of supraventricular tachycardia caused by a re-entry circuit within the right atrium",
            "fairly predictable atrial rate (NOT equal to ventricular rate for AFL) of around 300 bpm (range 200-400)",
            "fixed AV blocks, with ventricular rate a fraction (1/2,1/3,etc.) of atrial rate",
            "narrow complex tachycardia (ref. supraventricular & ventricular rate)",
            "flutter waves ('saw-tooth' pattern) best seen in leads II, III, aVF (may be more easily spotted by turning the ECG upside down), may resemble P waves in V1",
            "loss of the isoelectric baseline",  # important
        ],
    }
)

Brady = CFG(
    {  # rr
        "fullname": "bradycardia",
        "url": [
            "https://litfl.com/bradycardia-ddx/",
            "https://en.wikipedia.org/wiki/Bradycardia",
        ],
        "knowledge": [
            "heart rate (ventricular rate) < 60/min in an adult",
        ],
    }
)

BBB = CFG(
    {  # morphology
        "fullname": "bundle branch block",
        "url": [],
        "knowledge": [],
    }
)

IAVB = CFG(
    {  # morphology
        "fullname": "1st degree av block",
        "url": [
            "https://litfl.com/first-degree-heart-block-ecg-library/",
            "https://en.wikipedia.org/wiki/Atrioventricular_block#First-degree_Atrioventricular_Block",
        ],
        "knowledge": [
            "PR interval > 200ms",
            "Marked’ first degree block if PR interval > 300ms",
            "P waves might be buried in the preceding T wave",
            "there are no dropped, or skipped, beats",
        ],
    }
)

LBBB = CFG(
    {  # morphology
        "fullname": "left bundle branch block",
        "url": [
            "https://litfl.com/left-bundle-branch-block-lbbb-ecg-library/",
            "https://en.wikipedia.org/wiki/Left_bundle_branch_block",
        ],
        "knowledge": [
            "heart rhythm must be supraventricular",
            "QRS duration of > 120 ms",
            "lead V1: Dominant S wave, with QS or rS complex",
            "lateral leads: M-shaped, or notched, or broad monophasic R wave or RS complex; absence of Q waves (small Q waves are still allowed in aVL)",
            "chest (precordial) leads: poor R wave progression",
            "left precordial leads (V5-6): prolonged R wave peak time > 60ms",
            "ST segments and T waves always go in the opposite direction to the main vector of the QRS complex",
        ],
    }
)

CLBBB = CFG(
    {  # morphology
        "fullname": "complete left bundle branch block",
        "url": [],
        "knowledge": [],
    }
)

RBBB = CFG(
    {  # morphology
        "fullname": "right bundle branch block",
        "url": [
            "https://litfl.com/right-bundle-branch-block-rbbb-ecg-library/",
            "https://en.wikipedia.org/wiki/Right_bundle_branch_block",
        ],
        "knowledge": [
            "broad QRS > 100 ms (incomplete block) or > 120 ms (complete block)",
            "leads V1-3: RSR’ pattern (‘M-shaped’ QRS complex); sometimes a broad monophasic R wave or a qR complex in V1",
            "lateral leads: wide, slurred S wave",
        ],
    }
)

CRBBB = CFG(
    {  # morphology
        "fullname": "complete right bundle branch block",
        "url": [
            "https://litfl.com/right-bundle-branch-block-rbbb-ecg-library/",
            "https://en.wikipedia.org/wiki/Right_bundle_branch_block",
        ],
        "knowledge": [
            "broad QRS > 120 ms",
            "leads V1-3: RSR’ pattern (‘M-shaped’ QRS complex); sometimes a broad monophasic R wave or a qR complex in V1",
            "lateral leads: wide, slurred S wave",
        ],
    }
)

IRBBB = CFG(
    {  # morphology
        "fullname": "incomplete right bundle branch block",
        "url": [
            "https://litfl.com/right-bundle-branch-block-rbbb-ecg-library/",
            "https://en.wikipedia.org/wiki/Right_bundle_branch_block#Diagnosis",
        ],
        "knowledge": [
            "defined as an RSR’ pattern in V1-3 with QRS duration < 120ms (and > 100ms?)",
            "normal variant, commonly seen in children (of no clinical significance)",
        ],
    }
)

LAnFB = CFG(
    {  # morphology
        "fullname": "left anterior fascicular block",
        "url": [
            "https://litfl.com/left-anterior-fascicular-block-lafb-ecg-library/",
            "https://en.wikipedia.org/wiki/Left_anterior_fascicular_block",
        ],
        "knowledge": [
            "inferior leads (II, III, aVF): small R waves, large negative voltages (deep S waves), i.e. 'rS complexes'",
            "left-sided leads (I, aVL): small Q waves, large positive voltages (tall R waves), i.e. 'qR complexes'",
            "slight widening of the QRS",
            "increased R wave peak time in aVL",
            "LAD of degree (-45°, -90°)",
        ],
    }
)

LQRSV = CFG(
    {  # voltage
        "fullname": "low qrs voltages",
        "url": [
            "https://litfl.com/low-qrs-voltage-ecg-library/",
            "https://www.healio.com/cardiology/learn-the-heart/ecg-review/ecg-topic-reviews-and-criteria/low-voltage-review",
        ],
        "knowledge": [
            "peak-to-peak (VERY IMPORTANT) amplitudes of all the QRS complexes in the limb leads are < 5mm (0.5mV); or amplitudes of all the QRS complexes in the precordial leads are < 10mm (1mV)",
        ],
    }
)

PRWP = CFG(
    {  # voltage
        "fullname": "poor R wave progression",
        "url": [
            "https://litfl.com/poor-r-wave-progression-prwp-ecg-library/",
            "https://www.healio.com/cardiology/learn-the-heart/ecg-review/ecg-topic-reviews-and-criteria/poor-r-wave-progression",
            "https://emergencymedicinecases.com/ecg-cases-poor-r-wave-progression-late-mnemonic/",
            "https://www.wikidoc.org/index.php/Poor_R_Wave_Progression",
        ],
        "knowledge": [  # The definition of poor R wave progression (PRWP) varies in the literature
            "absence of the normal increase in size of the R wave in the precordial leads when advancing from lead V1 to V6",
            "in lead V1, the R wave should be small. The R wave becomes larger throughout the precordial leads, to the point where the R wave is larger than the S wave in lead V4. The S wave then becomes quite small in lead V6.",
            "failure of the R wave to progress in amplitude (R<3mm in V3), reversal of the progression (eg R in V2>V3), or delayed transition beyond V4",
            "R wave is less than 2–4 mm in leads V3 or V4 and/or there is presence of a reversed R wave progression, which is defined as R in V4 < R in V3 or R in V3 < R in V2 or R in V2 < R in V1, or any combination of these",
        ],
    }
)

NSIVCB = CFG(
    {  # mophology
        "fullname": "nonspecific intraventricular conduction disorder",
        "url": [
            "https://ecgwaves.com/topic/nonspecific-intraventricular-conduction-delay-defect/",
            "https://www.dynamed.com/condition/intraventricular-conduction-disorders-including-left-and-right-bundle-branch-block-lbbb-and-rbbb",
            "https://www.sciencedirect.com/science/article/pii/S0735109708041351",
            "https://www.heartrhythmjournal.com/article/S1547-5271(15)00073-9/abstract",
        ],
        "knowledge": [
            "widended (> 110ms) QRS complex, with not meeting the criteria (morphology different from) for LBBB and RBBB",
        ],
    }
)

PR = CFG(
    {  # morphology
        "fullname": "pacing rhythm",
        "url": [
            "https://litfl.com/pacemaker-rhythms-normal-patterns/",
            "https://www.youtube.com/watch?v=YkB4oX_COi8",
        ],
        "knowledge": [
            "morphology is dependent on the pacing mode (AAI, VVI, DDD, Magnet) used",
            "there are pacing spikes: vertical spikes of short duration, usually 2 ms (in doubt, viewing signals of the CINC2020 dataset, probably 10-20ms(at most half a small grid)?)",  # important
            "AAI (atrial pacing): pacing spike precedes the p wave",
            "VVI (ventricle pacing): pacing spike precedes the QRS complex; morphology similar to LBBB or RBBB (depending on lead placement)",
        ],
    }
)

PAC = CFG(
    {  # morphology, very complicated
        "fullname": "premature atrial contraction",
        "url": [
            "https://litfl.com/premature-atrial-complex-pac/",
            "https://en.wikipedia.org/wiki/Premature_atrial_contraction",
        ],
        "knowledge": [
            "an abnormal (non-sinus) P wave is followed by a QRS complex",
            "P wave typically has a different morphology and axis to the sinus P waves",
            "abnormal P wave may be hidden in the preceding T wave, producing a “peaked” or “camel hump” appearance",
            # to add more
        ],
    }
)

PJC = CFG(
    {  # morphology
        "fullname": "premature junctional contraction",
        "url": [
            "https://litfl.com/premature-junctional-complex-pjc/",
            "https://en.wikipedia.org/wiki/Premature_junctional_contraction",
        ],
        "knowledge": [
            "narrow QRS complex, either (1) without a preceding P wave or (2) with a retrograde P wave which may appear before, during, or after the QRS complex. If before, there is a short PR interval of < 120 ms and the  “retrograde” P waves are usually inverted in leads II, III and aVF",
            "occurs sooner than would be expected for the next sinus impulse",
            "followed by a compensatory pause",
        ],
    }
)

PVC = CFG(
    {  # morphology
        "fullname": "premature ventricular contractions",
        "url": [
            "https://litfl.com/premature-ventricular-complex-pvc-ecg-library/",
            "https://en.wikipedia.org/wiki/Premature_ventricular_contraction",
        ],
        "knowledge": [
            "broad QRS complex (≥ 120 ms) with abnormal morphology",
            "premature — i.e. occurs earlier than would be expected for the next sinus impulse",
            "discordant ST segment and T wave changes",
            "usually followed by a full compensatory pause",
            "retrograde capture of the atria may or may not occur",
        ],
    }
)

LPR = CFG(
    {  # morphology
        "fullname": "prolonged pr interval",
        "url": [
            "https://litfl.com/pr-interval-ecg-library/",
            "https://en.wikipedia.org/wiki/PR_interval",
            "https://www.healio.com/cardiology/learn-the-heart/ecg-review/ecg-interpretation-tutorial/pr-interval",
        ],
        "knowledge": [
            "PR interval >200ms",
        ],
    }
)

LQT = CFG(
    {  # morphology
        "fullname": "prolonged qt interval",
        "url": [
            "https://litfl.com/qt-interval-ecg-library/",
            "https://en.wikipedia.org/wiki/Long_QT_syndrome",
        ],
        "knowledge": [
            "LQT is measured by QTc (see ref url)"
            "QTc is prolonged if > 440ms in men or > 460ms in women (or > 480ms?)",
        ],
    }
)

QAb = CFG(
    {  # morphology
        "fullname": "qwave abnormal",
        "url": [
            "https://litfl.com/q-wave-ecg-library/",
            "https://en.ecgpedia.org/wiki/Pathologic_Q_Waves",
            "https://wikem.org/wiki/Pathologic_Q_waves",
        ],
        "knowledge": [
            "> 40 ms (1 mm) wide; > 2 mm deep; > 1/4 of depth of QRS complex in ANY TWO leads of a contiguous lead grouping: I, aVL,V6; V4–V6; II, III, aVF",
            "seen (≥ 0.02 s or QS complex) in leads V1-3",
        ],
    }
)

RAD = CFG(
    {  # morphology
        "fullname": "right axis deviation",
        "url": [
            "https://litfl.com/right-axis-deviation-rad-ecg-library/",
            "https://en.wikipedia.org/wiki/Right_axis_deviation",
            "https://www.ncbi.nlm.nih.gov/books/NBK470532/",
            "https://ecglibrary.com/axis.html",
            "https://www.youtube.com/watch?v=PbzDN2_rAFc",
        ],
        "knowledge": [  # should combine with LAD to study
            "QRS axis greater than +90°",
            "2-lead method: lead I is NEGATIVE; lead aVF is POSITIVE",  # important, ref LAD
            "3-lead method: lead I is NEGATIVE; lead aVF is POSITIVE (and II, III)",  # important, ref LAD
        ],
    }
)

LAD = CFG(
    {  # morphology
        "fullname": "left axis deviation",
        "url": [
            "https://litfl.com/left-axis-deviation-lad-ecg-library/",
            "https://en.wikipedia.org/wiki/Left_axis_deviation",
            "https://www.ncbi.nlm.nih.gov/books/NBK470532/",
            "https://ecglibrary.com/axis.html",
            "https://www.youtube.com/watch?v=PbzDN2_rAFc",
        ],
        "knowledge": [  # should combine with RAD to study
            "QRS axis (-30°, -90°)",
            "2-lead method: lead I is POSITIVE; lead aVF is NEGATIVE",  # important, ref RAD
            "3-lead method: POSITIVE in leads I (and aVL?); NEGATIVE in leads II, aVF, (and III?)",  # important, ref RAD
            "LAnFB, LBBB, PR, ventricular ectopics are causes of LAD",
        ],
    }
)

SA = CFG(
    {  # morphology
        "fullname": "sinus arrhythmia",
        "url": [
            "https://litfl.com/sinus-arrhythmia-ecg-library/",
            "https://www.healthline.com/health/sinus-arrhythmia",
        ],
        "knowledge": [
            "sinus rhythm (NSR), with a beat-to-beat variation (more than 120 ms) in the PP interval, producing an irregular ventricular rate",
            "PP interval gradually lengthens and shortens in a cyclical fashion, usually corresponding to the phases of the respiratory cycle",
            "normal sinus P waves (upright in leads I and II) with a constant morphology (i.e. not PAC)",
            "constant PR interval (i.e. not Mobitz I AV block)",
        ],
    }
)

SB = CFG(
    {  # rr, morphology
        "fullname": "sinus bradycardia",
        "url": [
            "https://litfl.com/sinus-bradycardia-ecg-library/",
            "https://en.wikipedia.org/wiki/Sinus_bradycardia",
        ],
        "knowledge": [
            "sinus rhythm (NSR); with a resting heart rate of < 60 bpm in adults (Brady)",
        ],
    }
)

NSR = CFG(
    {  # rr, morphology
        "fullname": "sinus rhythm",  # the NORMAL rhythm
        "url": [
            "https://litfl.com/normal-sinus-rhythm-ecg-library/",
            "https://en.wikipedia.org/wiki/Sinus_rhythm",
        ],
        "knowledge": [
            "regular rhythm (< 0.16 s variation in the shortest and longest durations between successive P waves) at a rate of 60-100 bpm",
            "each QRS complex is preceded by a normal P wave (positive in lead I, lead II, and aVF; negative in lead aVR; any of biphasic (-/+), positive or negative in lead aVL; positive in all chest leads, except for V1 which may be biphasic (+/-))",
            "normal PR interval, QRS complex and QT interval",  # normal
        ],
    }
)

STach = CFG(
    {  # rr, morphology,
        "fullname": "sinus tachycardia",
        "url": [
            "https://litfl.com/sinus-tachycardia-ecg-library/",
            "https://litfl.com/tachycardia-ddx/",
            "https://en.wikipedia.org/wiki/Sinus_tachycardia",
        ],
        "knowledge": [
            "sinus rhythm (NSR), with a resting heart rate of > 100 bpm in adults",
        ],
    }
)

SVPB = CFG(
    {  # morphology, equiv. to PAC for CINC2020
        "fullname": "supraventricular premature beats",
        "url": [
            "https://en.wikipedia.org/wiki/Premature_atrial_contraction#Supraventricular_extrasystole",
        ],
        "knowledge": PAC["knowledge"] + PJC["knowledge"],
    }
)

TAb = CFG(
    {  # morphology
        "fullname": "t wave abnormal",
        "url": [
            "https://litfl.com/t-wave-ecg-library/",
            "https://en.wikipedia.org/wiki/T_wave",
        ],
        "knowledge": [
            "normal T wave: upright in all leads, except aVR, aVL, III and V1; amplitude < 5mm in limb leads, < 10mm in precordial leads; asymmetrical with a rounded peak",
            "abnormalities: peaked (amplitude) T waves; hyperacute T waves (broad, symmetrical, usually with increased amplitude); inverted T waves (TInv); biphasic T waves; ‘camel hump’ T waves; flattened T waves (± 0.1mV)",
        ],
    }
)

TInv = CFG(
    {  # morphology
        "fullname": "t wave inversion",
        "url": [
            "https://en.wikipedia.org/wiki/T_wave#Inverted_T_wave",
            "https://litfl.com/t-wave-ecg-library/",
        ],
        "knowledge": [
            "normal T wave should be upright (positive peak amplitude) in all leads, except aVR, aVL, III and V1",
        ],
    }
)

VPB = CFG(
    {
        "fullname": "ventricular premature beats",
        "url": PVC["url"],
        "knowledge": PVC["knowledge"],
    }
)

SPB = SVPB  # alias

STD = CFG(
    {
        "fullname": "st depression",
        "url": [
            "https://litfl.com/st-segment-ecg-library/",
            "https://ecgwaves.com/st-segment-normal-abnormal-depression-elevation-causes/",
            "https://en.wikipedia.org/wiki/ST_elevation",
        ],
        "knowledge": [
            "",
        ],
    }
)

STE = CFG(
    {
        "fullname": "st elevation",
        "url": [
            "https://litfl.com/st-segment-ecg-library/",
            "https://ecgwaves.com/st-segment-normal-abnormal-depression-elevation-causes/",
            "https://en.wikipedia.org/wiki/ST_depression",
        ],
        "knowledge": [
            "",
        ],
    }
)


_dx_mapping = pd.read_csv(
    StringIO(
        """Dx,SNOMEDCTCode,Abbreviation
atrial fibrillation,164889003,AF
atrial flutter,164890007,AFL
bundle branch block,6374002,BBB
bradycardia,426627000,Brady
complete left bundle branch block,733534002,CLBBB
complete right bundle branch block,713427006,CRBBB
1st degree av block,270492004,IAVB
incomplete right bundle branch block,713426002,IRBBB
left axis deviation,39732003,LAD
left anterior fascicular block,445118002,LAnFB
left bundle branch block,164909002,LBBB
low qrs voltages,251146004,LQRSV
nonspecific intraventricular conduction disorder,698252002,NSIVCB
sinus rhythm,426783006,NSR
premature atrial contraction,284470004,PAC
pacing rhythm,10370003,PR
poor R wave Progression,365413008,PRWP
premature ventricular contractions,427172004,PVC
prolonged pr interval,164947007,LPR
prolonged qt interval,111975006,LQT
qwave abnormal,164917005,QAb
right axis deviation,47665007,RAD
right bundle branch block,59118001,RBBB
sinus arrhythmia,427393009,SA
sinus bradycardia,426177001,SB
sinus tachycardia,427084000,STach
supraventricular premature beats,63593006,SVPB
t wave abnormal,164934002,TAb
t wave inversion,59931005,TInv
ventricular premature beats,17338001,VPB
accelerated atrial escape rhythm,233892002,AAR
abnormal QRS,164951009,abQRS
atrial escape beat,251187003,AED
accelerated idioventricular rhythm,61277005,AIVR
accelerated junctional rhythm,426664006,AJR
suspect arm ecg leads reversed,251139008,ALR
acute myocardial infarction,57054005,AMI
acute myocardial ischemia,413444003,AMIs
anterior ischemia,426434006,AnMIs
anterior myocardial infarction,54329005,AnMI
atrial bigeminy,251173003,AB
atrial fibrillation and flutter,195080001,AFAFL
atrial hypertrophy,195126007,AH
atrial pacing pattern,251268003,AP
atrial rhythm,106068003,ARH
atrial tachycardia,713422000,ATach
av block,233917008,AVB
atrioventricular dissociation,50799005,AVD
atrioventricular junctional rhythm,29320008,AVJR
atrioventricular  node reentrant tachycardia,251166008,AVNRT
atrioventricular reentrant tachycardia,233897008,AVRT
blocked premature atrial contraction,251170000,BPAC
brugada,418818005,BRU
brady tachy syndrome,74615001,BTS
chronic atrial fibrillation,426749004,CAF
countercolockwise rotation,251199005,CCR
clockwise or counterclockwise vectorcardiographic loop,61721007,CVCL/CCVCL
cardiac dysrhythmia,698247007,CD
complete heart block,27885002,CHB
congenital incomplete atrioventricular heart block,204384007,CIAHB
coronary heart disease,53741008,CHD
chronic myocardial ischemia,413844008,CMI
clockwise rotation,251198002,CR
diffuse intraventricular block,82226007,DIB
early repolarization,428417006,ERe
fusion beats,13640000,FB
fqrs wave,164942001,FQRS
heart failure,84114007,HF
heart valve disorder,368009,HVD
high t-voltage,251259000,HTV
indeterminate cardiac axis,251200008,ICA
2nd degree av block,195042002,IIAVB
mobitz type II atrioventricular block,426183003,IIAVBII
inferior ischaemia,425419005,IIs
incomplete left bundle branch block,251120003,ILBBB
inferior ST segment depression,704997005,ISTD
idioventricular rhythm,49260003,IR
junctional escape,426995002,JE
junctional premature complex,251164006,JPC
junctional tachycardia,426648003,JTach
left atrial abnormality,253352002,LAA
left atrial enlargement,67741000119109,LAE
left atrial hypertrophy,446813000,LAH
lateral ischaemia,425623009,LIs
left posterior fascicular block,445211001,LPFB
left ventricular hypertrophy,164873001,LVH
left ventricular high voltage,55827005,LVHV
left ventricular strain,370365005,LVS
myocardial infarction,164865005,MI
myocardial ischemia,164861001,MIs
mobitz type i wenckebach atrioventricular block,54016002,MoI
nonspecific st t abnormality,428750005,NSSTTA
old myocardial infarction,164867002,OldMI
paroxysmal atrial fibrillation,282825002,PAF
prolonged P wave,251205003,PPW
paroxysmal supraventricular tachycardia,67198005,PSVT
paroxysmal ventricular tachycardia,425856008,PVT
p wave change,164912004,PWC
right atrial abnormality,253339007,RAAb
r wave abnormal,164921003,RAb
right atrial hypertrophy,446358003,RAH
right atrial  high voltage,67751000119106,RAHV
rapid atrial fibrillation,314208002,RAF
right ventricular hypertrophy,89792004,RVH
sinus atrium to atrial wandering rhythm,17366009,SAAWR
sinoatrial block,65778007,SAB
sinus arrest,5609005,SARR
sinus node dysfunction,60423000,SND
shortened pr interval,49578007,SPRI
decreased qt interval,77867006,SQT
s t changes,55930002,STC
st depression,429622005,STD
st elevation,164931005,STE
st interval abnormal,164930006,STIAb
supraventricular bigeminy,251168009,SVB
supraventricular tachycardia,426761007,SVT
transient ischemic attack,266257000,TIA
tall p wave,251223006,TPW
u wave abnormal,164937009,UAb
ventricular bigeminy,11157007,VBig
ventricular ectopics,164884008,VEB
ventricular escape beat,75532003,VEsB
ventricular escape rhythm,81898007,VEsR
ventricular fibrillation,164896001,VF
ventricular flutter,111288001,VFL
ventricular hypertrophy,266249003,VH
ventricular pre excitation,195060002,VPEx
ventricular pacing pattern,251266004,VPP
paired ventricular premature complexes,251182009,VPVC
ventricular tachycardia,164895002,VTach
ventricular trigeminy,251180001,VTrig
wandering atrial pacemaker,195101003,WAP
wolff parkinson white pattern,74390002,WPW"""
    )
)


for ea_str in __all__:
    ea = eval(ea_str)
    try:
        ea["url"].insert(
            0,
            f"{_snomedbrowser_url}{_dx_mapping[_dx_mapping.Abbreviation==ea_str]['SNOMEDCTCode'].values[0]}",
        )
    except Exception:
        pass



"""
use median filter to remove baseline,
note that highpass filters also have the effect of baseline removal
"""

import warnings
from numbers import Real
from typing import Any, List, NoReturn, Tuple

import numpy as np

from .base import PreProcessor, preprocess_multi_lead_signal

__all__ = [
    "BaselineRemove",
]


class BaselineRemove(PreProcessor):
    """ """

    __name__ = "BaselineRemove"

    def __init__(
        self, window1: float = 0.2, window2: float = 0.6, **kwargs: Any
    ) -> NoReturn:
        """

        Parameters
        ----------
        window1: float, default 0.2,
            the smaller window size of the median filter, with units in seconds
        highcut: float, default 0.6,
            the larger window size of the median filter, with units in seconds
        """
        self.window1 = window1
        self.window2 = window2
        if self.window2 < self.window1:
            self.window1, self.window2 = self.window2, self.window1
            warnings.warn("values of window1 and window2 are switched")

    def apply(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        """

        apply the preprocessor to `sig`

        Parameters
        ----------
        sig: ndarray,
            the ECG signal, can be
            1d array, which is a single-lead ECG
            2d array, which is a multi-lead ECG of "lead_first" format
            3d array, which is a tensor of several ECGs, of shape (batch, lead, siglen)
        fs: real number,
            sampling frequency of the ECG signal

        Returns
        -------
        filtered_sig: ndarray,
            the median filtered (hence baseline removed) ECG signal
        fs: int,
            the sampling frequency of the filtered ECG signal
        """
        self._check_sig(sig)
        filtered_sig = preprocess_multi_lead_signal(
            raw_sig=sig,
            fs=fs,
            bl_win=[self.window1, self.window2],
        )
        return filtered_sig, fs

    def extra_repr_keys(self) -> List[str]:
        """
        return the extra keys for `__repr__`
        """
        return ["window1", "window2"] + super().extra_repr_keys()



# -*- coding: utf-8 -*-
"""
"""

from numbers import Real
from pathlib import Path
from typing import Any, Dict, List, NoReturn, Optional, Sequence, Union  # noqa: F401

import numpy as np
import pandas as pd  # noqa: F401
import wfdb
from scipy.signal import resample_poly

from ...cfg import CFG, DEFAULTS
from ...utils.misc import add_docstring
from ...utils.utils_data import ECGWaveForm
from ..base import (
    PhysioNetDataBase,
    DataBaseInfo,
    BeatAnn,
    WFDB_Beat_Annotations,
    WFDB_Non_Beat_Annotations,
)


__all__ = [
    "QTDB",
]


_QTDB_INFO = DataBaseInfo(
    title="""
    QT Database
    """,
    about="""
    1. The QT Database includes ECGs which were chosen to represent a wide variety of QRS and ST-T morphologies
    2. Recordings were chosen chosen from the MIT-BIH Arrhythmia Database (MITDB), the European Society of Cardiology ST-T Database (EDB), and several other ECG databases collected at Boston's Beth Israel Deaconess Medical Center (MIT-BIH ST Change Database, MIT-BIH Supraventricular Arrhythmia Database, MIT-BIH Normal Sinus Rhythm Database, MIT-BIH Long-Term ECG Database, ``sudden death'' patients from BIH)
    2. Contains 105 fifteen-minute two-lead ECG recordings
    3. Contains onset, peak, and end markers for P, QRS, T, and (where present) U waves of from 30 to 50 selected beats in each recording
    4. Annotation file table:
        | Suffix | Meaning |
        | ------ | ------- |
        | .atr   | reference beat annotations from original database (not available for the 24 sudden death records) |
        | .man:  | reference beat annotations for selected beats only |
        | .q1c:  | manually determined waveform boundary measurements for selected beats (annotator 1 only -- second pass) |
        | .q2c:  | manually determined waveform boundary measurements for selected beats (annotator 2 only -- second pass; available for only 11 records) |
        | .q1t:  | manually determined waveform boundary measurements for selected beats (annotator 1 only -- first pass) |
        | .q2t:  | manually determined waveform boundary measurements for selected beats (annotator 2 only -- first pass; available for only 11 records) |
        | .pu:   | automatically determined waveform boundary measurements for all beats (based on both signals) |
        | .pu0:  | automatically determined waveform boundary measurements for all beats (based on signal 0 only) |
        | .pu1:  | automatically determined waveform boundary measurements for all beats (based on signal 1 only) |
    5. A part of the recordings have rhythm annotations, ST change (elevation or depression) annotations, all of which have .atr annotation files. These annotations are provided in the `aux_note` attribute of the annotation object.
    6. In the first pass manual wave delineation annotation files (.qt1, .qt2 files), fiducial points were marked by a "|" symbol, along with beat annotations (one of "A", "B", "N", "Q", "V") inherited from corresponding .man files.
    7. In the second pass manual wave delineation annotation files (.q1c, .q2c files), the final manual annotations are recorded, with the regular annotation symbols "(" ,")", "t", "p", and "u", and with annotations inherited from the .qt1, .qt2 files.
    8. The .pu0, .pu1 files contain the automatic waveform onsets and ends in signals 0 and 1 respectively, as detected using the differentiated threshold method by ecgpuwave. In the num fields of the pu* annotations, ecgpuwave classifies the T waves as normal (0), inverted (1), only upwards (2), only downwards (3), biphasic negative-positive (4), or biphasic positive-negative (5). Waveform onset ``('' and offset ``)'' annotations specify the waveform type in their num fields (0 for a P-wave, 1 for a QRS complex, 2 for a T wave, or 3 for a U-wave).
    """,
    usage=[
        "ECG wave delineation",
        "ST segment",
    ],
    references=[
        "https://physionet.org/content/qtdb/1.0.0/",
        "Laguna P, Mark RG, Goldberger AL, Moody GB. A Database for Evaluation of Algorithms for Measurement of QT and Other Waveform Intervals in the ECG. Computers in Cardiology 24:673-676 (1997).",
    ],
    issues="""
    1. According to the paper of the database, there should be .ari files containing QRS annotations obtained automatically by ARISTOTLE, which however are not available in the database.
    2. A large proportion of the wave delineation annotations lack onset indices (the T waves and U waves).
    """,
    doi=[
        "10.1109/cic.1997.648140",
        "10.13026/C24K53",
    ],
)


@add_docstring(_QTDB_INFO.format_database_docstring())
class QTDB(PhysioNetDataBase):
    """ """

    __name__ = "QTDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 2,
        **kwargs: Any,
    ) -> NoReturn:
        """

        Parameters
        ----------
        db_dir: str or Path, optional,
            storage path of the database
        working_dir: str or Path, optional,
            working directory, to store intermediate files and log file
        verbose: int, default 2,
            log verbosity
        kwargs: auxilliary key word arguments

        """
        super().__init__(
            db_name="qtdb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 250
        self.spacing = 1000 / self.fs
        self.data_ext = "dat"

        # fmt: off
        self.all_extensions = ["atr", "man", "q1c", "q2c", "qt1", "qt2", "pu", "pu0", "pu1"]
        """
        1. .atr:    reference beat annotations from original database (not available in all cases)
        2. .man:    reference beat annotations for selected beats only
        3. .q1c:    manually determined waveform boundary measurements for selected beats (annotator 1 only -- second pass)
        4. .q2c:    manually determined waveform boundary measurements for selected beats (annotator 2 only -- second pass; available for only 11 records)
        5. .q1t:    manually determined waveform boundary measurements for selected beats (annotator 1 only -- first pass)
        6. .q2t:    manually determined waveform boundary measurements for selected beats (annotator 2 only -- first pass; available for only 11 records)
        7. .pu:     automatically determined waveform boundary measurements for all beats (based on both signals)
        8. .pu0:    automatically determined waveform boundary measurements for all beats (based on signal 0 only)
        9. .pu1:    automatically determined waveform boundary measurements for all beats (based on signal 1 only)
        """
        self.all_leads = [
            "CC5", "CM2", "CM4", "CM5", "D3", "D4", "ECG1", "ECG2", "ML5", "MLII",
            "V1", "V1-V2", "V2", "V2-V3", "V3", "V4", "V4-V5", "V5", "mod.V1",
        ]
        self.all_annotations = ["(", ")", "N", "t", "p"]
        # fmt: on

        self.beat_types_extended = list("""~"+/AFJNQRSTVaefjs|""")
        self.nonbeat_types = [
            item
            for item in self.beat_types_extended
            if item in WFDB_Non_Beat_Annotations
        ]
        self.beat_types = [
            item for item in self.beat_types_extended if item in WFDB_Beat_Annotations
        ]
        self.beat_types_map = {item: i for i, item in enumerate(self.beat_types)}
        self.beat_types_extended_map = {
            item: i for i, item in enumerate(self.beat_types_extended)
        }

        self.class_map = CFG(p=1, N=2, t=3, i=0)  # an extra isoelectric

        self._ls_rec()

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`

        Returns
        -------
        sid: int,
            the `subject_id` corr. to `rec`

        """
        raise NotImplementedError

    def get_lead_names(self, rec: Union[str, int]) -> List[str]:
        """
        get the lead names of the record `rec`

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`

        Returns
        -------
        list of str,
            list of the lead names

        """
        return wfdb.rdheader(str(self.get_absolute_path(rec))).sig_name

    def _normalize_leads(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, int, List[str], List[int]]] = None,
    ) -> List[str]:
        """
        normalize the lead names to the ones in `self.all_leads`

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        leads: str or int or list of str or list of int,
            lead names or indices of the leads in `self.all_leads`

        Returns
        -------
        list of str,
            list of the normalized lead names

        """
        if leads is not None:
            if isinstance(leads, str):
                leads = [leads]
            elif isinstance(leads, int):
                leads = [self.get_lead_names(rec)[leads]]
            elif isinstance(leads, (list, tuple)):
                if isinstance(leads[0], int):
                    leads = [self.get_lead_names(rec)[ld] for ld in leads]
                if isinstance(leads[0], str):
                    assert set(leads) <= set(
                        self.get_lead_names(rec)
                    ), f"leads must be a subset of the record's total leads {self.get_lead_names(rec)}"
                else:
                    raise ValueError(
                        "leads must be str or int or a list of str or int, "
                        f"but got {type(leads[0])}"
                    )
            else:
                raise ValueError(
                    "leads must be str or int or a list of str or int, "
                    f"but got {type(leads)}"
                )
        else:
            leads = self.get_lead_names(rec)
        return leads

    def load_data(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, int, List[str], List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format="channel_first",
        units: str = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """
        load physical (converted from digital) ECG data,
        which is more understandable for humans

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        leads: str or int or list of str or list of int, optional,
            the leads to load
        sampfrom: int, optional,
            start index of the data to be loaded
        sampto: int, optional,
            end index of the data to be loaded
        data_format: str, default "channel_first",
            format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first", original)
        units: str, default "mV",
            units of the output signal, can also be "μV", with an alias of "uV"
        fs: real number, optional,
            if not None, the loaded data will be resampled to this frequency

        Returns
        -------
        data: ndarray,
            the ECG data

        """
        if isinstance(rec, int):
            rec = self[rec]
        assert data_format.lower() in [
            "channel_first",
            "lead_first",
            "channel_last",
            "lead_last",
        ]
        rec_fp = self.get_absolute_path(rec)
        leads = self._normalize_leads(rec, leads)

        wfdb_rec = wfdb.rdrecord(
            str(rec_fp),
            sampfrom=sampfrom or 0,
            sampto=sampto,
            physical=True,
            channel_names=leads,
        )
        data = np.asarray(wfdb_rec.p_signal.T, dtype=DEFAULTS.np_dtype)

        if units.lower() in ["uv", "μv"]:
            data = data * 1000
        if fs is not None and fs != wfdb_rec.fs:
            data = resample_poly(data, fs, wfdb_rec.fs, axis=1).astype(
                DEFAULTS.np_dtype
            )
        if data_format.lower() in ["channel_last", "lead_last"]:
            data = data.T

        return data

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
        ignore_beat_types: bool = True,
        extension: str = "q1c",
    ) -> List[ECGWaveForm]:
        """
        load the wave delineation in the form of list of `ECGWaveForm`

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        ignore_beat_types: bool, default True,
            if True, ignore the beat types (all converted to "N") in the annotation file
        extension: str, default "q1c",
            the extension of the wave delineation file

        Returns
        -------
        wave_list: list of `ECGWaveForm`,
            the list of wave delineation in the form of `ECGWaveForm`

        """
        assert extension in ["q1c", "q2c", "pu1", "pu2"]
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=extension)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        subtraction = 0 if keep_original else sf

        wave_list = []
        current_onset = None
        current_wave_name = None
        current_wave_peak = None
        for idx, symbol in zip(wfdb_ann.sample, wfdb_ann.symbol):
            if idx < sf:
                continue
            if idx >= st:
                break
            if symbol == "(":
                current_onset = idx
            elif symbol == ")":
                wave_list.append(
                    ECGWaveForm(
                        onset=(current_onset or np.nan) - subtraction,
                        offset=idx - subtraction,
                        name=current_wave_name,
                        peak=current_wave_peak,
                        duration=(idx - current_onset) / header.fs
                        if current_onset is not None
                        else np.nan,
                    )
                )
                current_onset = None
                current_wave_name = None
                current_wave_peak = None
            else:
                if ignore_beat_types and symbol not in ["p", "t", "u"]:
                    symbol = "N"
                current_wave_name = symbol
                current_wave_peak = idx

        return wave_list

    @add_docstring(load_ann.__doc__)
    def load_wave_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
        ignore_beat_types: bool = True,
        extension: str = "q1c",
    ) -> np.ndarray:
        """alias of self.load_ann"""
        return self.load_ann(
            rec,
            sampfrom=sampfrom,
            sampto=sampto,
            keep_original=keep_original,
            ignore_beat_types=ignore_beat_types,
            extension=extension,
        )

    def load_wave_masks(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        mask_format: str = "channel_first",
        class_map: Optional[Dict[str, int]] = None,
        extension: str = "q1c",
    ) -> np.ndarray:
        """
        load the wave delineation in the form of list of `ECGWaveForm`

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        mask_format: str, default "channel_first",
            format of the mask,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first")
        class_map: dict, optional,
            custom class map,
            if not set, `self.class_map` will be used
        extension: str, default "q1c",
            the extension of the wave delineation file

        Returns
        -------
        masks: ndarray,
            the masks corresponding to the wave delineation annotations of `rec`

        """
        raise NotImplementedError(
            "A large proportion of the wave delineation annotations lack onset indices. "
            "Has to find a rule to give default onset index for the missing ones."
        )

    def load_rhythm_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rhythm_format: str = "intervals",
        rhythm_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
        extension: str = "atr",
    ) -> Union[Dict[str, list], np.ndarray]:
        """
        load rhythm annotations,
        which are stored in the `aux_note` attribute of corresponding annotation files.

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        rhythm_format: str, default "intervals", case insensitive,
            format of returned annotation, can also be "mask"
        rhythm_types: list of str, optional,
            defaults to `self.rhythm_types`
            if not None, only the rhythm annotations with the specified types will be returned
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        extension: str, default "atr",
            the extension of the annotation file,
            has to be "atr", since "man" files has no rhythm annotation

        Returns
        -------
        ann, dict or ndarray,
            the annotations in the format of intervals, or in the format of mask

        """
        raise NotImplementedError(
            "only a small part of the recordings have rhythm annotations, "
            "hence not implemented yet"
        )

    def load_beat_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        beat_format: str = "beat",
        beat_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
        extension: str = "atr",
    ) -> Union[Dict[str, np.ndarray], List[BeatAnn]]:
        """
        load beat annotations,
        which are stored in the `symbol` attribute of corresponding annotation files

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        beat_format: str, default "beat", case insensitive,
            format of returned annotation, can also be "dict"
        beat_types: list of str, optional,
            defaults to `self.beat_types`
            if not None, only the beat annotations with the specified types will be returned
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        extension: str, default "atr",
            the extension of the annotation file, can also be "man"

        Returns
        -------
        beat_ann: dict or list,
            locations (indices) of the all the beat types

        """
        assert beat_format.lower() in [
            "beat",
            "dict",
        ], f"`beat_format` must be one of ['beat', 'dict'], got {beat_format}"
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=extension)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        sample_inds = wfdb_ann.sample
        indices = np.where((sample_inds >= sf) & (sample_inds < st))[0]

        if beat_types is None:
            beat_types = self.beat_types

        beat_ann = [
            BeatAnn(i, s)
            for i, s in zip(sample_inds[indices], np.array(wfdb_ann.symbol)[indices])
            if s in beat_types
        ]

        if beat_format.lower() == "dict":
            beat_ann = {
                s: np.array([b.index for b in beat_ann if b.symbol == s], dtype=int)
                for s in self.beat_types_extended
            }
            beat_ann = {k: v for k, v in beat_ann.items() if len(v) > 0}

        return beat_ann

    def load_rpeak_indices(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
        extension: str = "atr",
    ) -> np.ndarray:
        """
        load rpeak indices, or equivalently qrs complex locations,
        which are stored in the `symbol` attribute of corresponding annotation files,
        regardless of their beat types,

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        extension: str, default "atr",
            the extension of the annotation file, can also be "man"

        Returns
        -------
        rpeak_inds: ndarray,
            locations (indices) of the all the rpeaks (qrs complexes)

        """
        assert extension in ["atr", "man"]
        if isinstance(rec, int):
            rec = self[rec]
        rec_fp = self.get_absolute_path(rec)
        if not rec_fp.with_suffix(f".{extension}").exists():
            another_extension = "man" if extension == "atr" else "atr"
            raise FileNotFoundError(
                f"annotation file {rec_fp.name} does not exist, "
                f"try setting `extension = \042{another_extension}\042`"
            )
        wfdb_ann = wfdb.rdann(str(rec_fp), extension=extension)
        header = wfdb.rdheader(str(rec_fp))
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        rpeak_inds = wfdb_ann.sample
        indices = np.where(
            (rpeak_inds >= sf)
            & (rpeak_inds < st)
            & (np.isin(wfdb_ann.symbol, self.beat_types))
        )[0]
        rpeak_inds = rpeak_inds[indices]
        if not keep_original:
            rpeak_inds -= sf
        return rpeak_inds

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[str, int, List[str], List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        same_range: bool = False,
        waves: Optional[ECGWaveForm] = None,
        beat_ann: Optional[Dict[str, np.ndarray]] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
        **kwargs: Any,
    ) -> NoReturn:
        """
        plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        data: ndarray, optional,
            if given, data of `rec` will not be used,
            this is useful when plotting filtered data
        ticks_granularity: int, default 0,
            the granularity to plot axis ticks, the higher the more,
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads: str or int or list of str or list of int, optional,
            the leads to plot
        sampfrom: int, optional,
            start index of the record to plot
        sampto: int, optional,
            end index of the record to plot
        same_range: bool, default False,
            if True, forces all leads to have the same y range
        waves: ECGWaveForm, optional,
            the waves (p waves, t waves, qrs complexes, etc.)
        beat_ann: dict, optional,
            the beat annotations
        rpeak_inds: ndarray or list of int, optional,
            the rpeak indices
        kwargs: dict,

        TODO:
        -----
        1. slice too long records, and plot separately for each segment
        2. plot waves using `axvspan`

        NOTE
        ----
        `Locator` of `plt` has default `MAXTICKS` equal to 1000,
        if not modifying this number, at most 40 seconds of signal could be plotted once

        Contributors: Jeethan, and WEN Hao

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000
        _leads = self._normalize_leads(leads, standard_ordering=True, lower_cases=False)

        raise NotImplementedError



"""
"""

# import pytest

try:
    import torch_ecg  # noqa: F401
except ModuleNotFoundError:
    import sys
    from pathlib import Path

    sys.path.insert(0, str(Path(__file__).absolute().parents[1]))

# from torch_ecg.augmenters.augmenter_manager import AugmenterManager
# from torch_ecg.augmenters.baseline_wander import BaselineWanderAugmenter
# from torch_ecg.augmenters.label_smooth import LabelSmooth
# from torch_ecg.augmenters.mixup import Mixup
# from torch_ecg.augmenters.random_flip import RandomFlip
# from torch_ecg.augmenters.random_masking import RandomMasking
# from torch_ecg.augmenters.stretch_compress import StretchCompress



"""
PreProcessors as torch.nn.Module
"""

from .bandpass import BandPass
from .baseline_remove import BaselineRemove
from .normalize import MinMaxNormalize, NaiveNormalize, Normalize, ZScoreNormalize
from .preproc_manager import PreprocManager
from .resample import Resample

__all__ = [
    "PreprocManager",
    "BandPass",
    "BaselineRemove",
    "Normalize",
    "MinMaxNormalize",
    "NaiveNormalize",
    "ZScoreNormalize",
    "Resample",
]



"""
utilities for nn models

"""

import re
from copy import deepcopy
from itertools import repeat
from math import floor
from numbers import Real
from typing import List, Optional, Sequence, Tuple, Union, Dict

import numpy as np
import torch
from torch import Tensor, nn

from ..cfg import DEFAULTS

if DEFAULTS.torch_dtype == torch.float64:
    torch.set_default_tensor_type(torch.DoubleTensor)


__all__ = [
    "extend_predictions",
    "compute_output_shape",
    "compute_conv_output_shape",
    "compute_deconv_output_shape",
    "compute_maxpool_output_shape",
    "compute_avgpool_output_shape",
    "compute_sequential_output_shape",
    "compute_sequential_output_shape_docstring",
    "compute_module_size",
    "default_collate_fn",
    "compute_receptive_field",
    "adjust_cnn_filter_lengths",
    "SizeMixin",
    "CkptMixin",
]


def extend_predictions(
    preds: Sequence, classes: List[str], extended_classes: List[str]
) -> np.ndarray:
    """

    extend the prediction arrays to prediction arrays in larger range of classes

    Parameters
    ----------
    preds: sequence,
        sequence of predictions (scalar or binary),
        of shape (n_records, n_classes), or (n_classes,),
        where n_classes = `len(classes)`
    classes: list of str,
        classes of the predictions of `preds`
    extended_classes: list of str,
        a superset of `classes`

    Returns
    -------
    extended_preds: ndarray,
        the extended array of predictions, with indices in `extended_classes`,
        of shape (n_records, n_classes), or (n_classes,)

    """
    _preds = np.atleast_2d(preds)
    assert _preds.shape[1] == len(
        classes
    ), f"`pred` indicates {_preds.shape[1]} classes, while `classes` has {len(classes)}"
    assert (
        len(set(classes) - set(extended_classes)) == 0
    ), f"`extended_classes` is not a superset of `classes`, with {set(classes)-set(extended_classes)} in `classes` but not in `extended_classes`"

    extended_preds = np.zeros((_preds.shape[0], len(extended_classes)))

    for idx, c in enumerate(classes):
        new_idx = extended_classes.index(c)
        extended_preds[..., new_idx] = _preds[..., idx]

    if np.array(preds).ndim == 1:
        extended_preds = extended_preds[0]

    return extended_preds


# utils for computing output shape
def compute_output_shape(
    layer_type: str,
    input_shape: Sequence[Union[int, None]],
    num_filters: Optional[int] = None,
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    output_padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """

    compute the output shape of a (transpose) convolution/maxpool/avgpool layer

    Parameters
    ----------
    layer_type: str,
        type (conv, maxpool, avgpool, etc.) of the layer
    input_shape: sequence of int or None,
        shape of an input Tensor,
        the first dimension is the batch dimension, which is allowed to be `None`
    num_filters: int, optional,
        number of filters, also the channel dimension
    kernel_size: int, or sequence of int, default 1,
        kernel size (filter size) of the layer, should be compatible with `input_shape`
    stride: int, or sequence of int, default 1,
        stride (down-sampling length) of the layer, should be compatible with `input_shape`
    padding: int, or sequence of int, default 0,
        padding length(s) of the layer, should be compatible with `input_shape`
    out_padding: int, or sequence of int, default 0,
        additional size added to one side of the output shape,
        used only for transpose convolution
    dilation: int, or sequence of int, default 1,
        dilation of the layer, should be compatible with `input_shape`
    channel_last: bool, default False,
        channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention)

    Returns
    -------
    output_shape: tuple,
        shape of the output Tensor

    References
    ----------
    [1] https://discuss.pytorch.org/t/utility-function-for-calculating-the-shape-of-a-conv-output/11173/5

    """
    __TYPES__ = [
        "conv",
        "convolution",
        "deconv",
        "deconvolution",
        "transposeconv",
        "transposeconvolution",
        "maxpool",
        "maxpooling",
        "avgpool",
        "avgpooling",
        "averagepool",
        "averagepooling",
    ]
    lt = "".join(layer_type.lower().split("_"))
    assert lt in __TYPES__
    if lt in [
        "conv",
        "convolution",
    ]:
        # as function of dilation, kernel_size
        def minus_term(d, k):
            return d * (k - 1) + 1

        out_channels = num_filters
    elif lt in [
        "maxpool",
        "maxpooling",
    ]:

        def minus_term(d, k):
            return d * (k - 1) + 1

        out_channels = input_shape[-1] if channel_last else input_shape[1]
    elif lt in [
        "avgpool",
        "avgpooling",
        "averagepool",
        "averagepooling",
    ]:

        def minus_term(d, k):
            return k

        out_channels = input_shape[-1] if channel_last else input_shape[1]
    elif lt in [
        "deconv",
        "deconvolution",
        "transposeconv",
        "transposeconvolution",
    ]:
        out_channels = num_filters
    dim = len(input_shape) - 2
    assert (
        dim > 0
    ), "input_shape should be a sequence of length at least 3, to be a valid (with batch and channel) shape of a non-degenerate Tensor"

    # none_dim_msg = "only batch and channel dimension can be `None`"
    # if channel_last:
    #     assert all([n is not None for n in input_shape[1:-1]]), none_dim_msg
    # else:
    #     assert all([n is not None for n in input_shape[2:]]), none_dim_msg
    none_dim_msg = "spatial dimensions should be all `None`, or all not `None`"
    if channel_last:
        if all([n is None for n in input_shape[1:-1]]):
            if out_channels is None:
                raise ValueError(
                    "out channel dimension and spatial dimensions are all `None`"
                )
            output_shape = tuple(list(input_shape[:-1]) + [out_channels])
            return output_shape
        elif any([n is None for n in input_shape[1:-1]]):
            raise ValueError(none_dim_msg)
    else:
        if all([n is None for n in input_shape[2:]]):
            if out_channels is None:
                raise ValueError(
                    "out channel dimension and spatial dimensions are all `None`"
                )
            output_shape = tuple([input_shape[0], out_channels] + list(input_shape[2:]))
            return output_shape
        elif any([n is None for n in input_shape[2:]]):
            raise ValueError(none_dim_msg)

    if isinstance(kernel_size, int):
        _kernel_size = list(repeat(kernel_size, dim))
    elif len(kernel_size) == dim:
        _kernel_size = kernel_size
    else:
        raise ValueError(
            f"input has {dim} dimensions, while kernel has {len(kernel_size)} dimensions, both not including the channel dimension"
        )

    if isinstance(stride, int):
        _stride = list(repeat(stride, dim))
    elif len(stride) == dim:
        _stride = stride
    else:
        raise ValueError(
            f"input has {dim} dimensions, while kernel has {len(stride)} dimensions, both not including the channel dimension"
        )

    if isinstance(padding, int):
        _padding = list(repeat(list(repeat(padding, 2)), dim))
    elif len(padding) == 2 and isinstance(padding[0], int):
        _padding = list(repeat(padding, dim))
    elif (
        len(padding) == dim
        and all([isinstance(p, Sequence) for p in padding])
        and all([len(p) == 2 for p in padding])
    ):
        _padding = padding
    else:
        # raise ValueError(
        #     f"input has {dim} dimensions, while kernel has {len(padding)} dimensions, both not including the channel dimension"
        # )
        raise ValueError("Invalid padding")

    if isinstance(output_padding, int):
        _output_padding = list(repeat(output_padding, dim))
    elif len(output_padding) == dim:
        _output_padding = output_padding
    else:
        raise ValueError(
            f"input has {dim} dimensions, while kernel has {len(output_padding)} dimensions, both not including the channel dimension"
        )

    if isinstance(dilation, int):
        _dilation = list(repeat(dilation, dim))
    elif len(dilation) == dim:
        _dilation = dilation
    else:
        raise ValueError(
            f"input has {dim} dimensions, while kernel has {len(dilation)} dimensions, both not including the channel dimension"
        )

    if channel_last:
        _input_shape = list(input_shape[1:-1])
    else:
        _input_shape = list(input_shape[2:])

    if lt in [
        "deconv",
        "deconvolution",
        "transposeconv",
        "transposeconvolution",
    ]:
        output_shape = [
            (i - 1) * s - sum(p) + d * (k - 1) + o + 1
            for i, p, o, d, k, s in zip(
                _input_shape,
                _padding,
                _output_padding,
                _dilation,
                _kernel_size,
                _stride,
            )
        ]
    else:
        output_shape = [
            floor(((i + sum(p) - minus_term(d, k)) / s) + 1)
            for i, p, d, k, s in zip(
                _input_shape, _padding, _dilation, _kernel_size, _stride
            )
        ]
    if channel_last:
        output_shape = tuple([input_shape[0]] + output_shape + [out_channels])
    else:
        output_shape = tuple([input_shape[0], out_channels] + output_shape)

    return output_shape


def compute_conv_output_shape(
    input_shape: Sequence[Union[int, None]],
    num_filters: Optional[int] = None,
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """finished, cheched,

    compute the output shape of a convolution/maxpool/avgpool layer

    input_shape: sequence of int or None,
        shape of an input Tensor,
        the first dimension is the batch dimension, which is allowed to be `None`
    num_filters: int, optional,
        number of filters, also the channel dimension
    kernel_size: int, or sequence of int, default 1,
        kernel size (filter size) of the layer, should be compatible with `input_shape`
    stride: int, or sequence of int, default 1,
        stride (down-sampling length) of the layer, should be compatible with `input_shape`
    padding: int, or sequence of int, default 0,
        padding length(s) of the layer, should be compatible with `input_shape`
    dilation: int, or sequence of int, default 1,
        dilation of the layer, should be compatible with `input_shape`
    channel_last: bool, default False,
        channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention)

    Returns
    -------
    output_shape: tuple,
        shape of the output Tensor

    """
    output_shape = compute_output_shape(
        "conv",
        input_shape,
        num_filters,
        kernel_size,
        stride,
        padding,
        0,
        dilation,
        channel_last,
    )
    return output_shape


def compute_maxpool_output_shape(
    input_shape: Sequence[Union[int, None]],
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """finished, cheched,

    compute the output shape of a maxpool layer

    input_shape: sequence of int or None,
        shape of an input Tensor,
        the first dimension is the batch dimension, which is allowed to be `None`
    kernel_size: int, or sequence of int, default 1,
        kernel size (filter size) of the layer, should be compatible with `input_shape`
    stride: int, or sequence of int, default 1,
        stride (down-sampling length) of the layer, should be compatible with `input_shape`
    padding: int, or sequence of int, default 0,
        padding length(s) of the layer, should be compatible with `input_shape`
    dilation: int, or sequence of int, default 1,
        dilation of the layer, should be compatible with `input_shape`
    channel_last: bool, default False,
        channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention)

    Returns
    -------
    output_shape: tuple,
        shape of the output Tensor

    """
    output_shape = compute_output_shape(
        "maxpool",
        input_shape,
        1,
        kernel_size,
        stride,
        padding,
        0,
        dilation,
        channel_last,
    )
    return output_shape


def compute_avgpool_output_shape(
    input_shape: Sequence[Union[int, None]],
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """finished, cheched,

    compute the output shape of a avgpool layer

    input_shape: sequence of int or None,
        shape of an input Tensor,
        the first dimension is the batch dimension, which is allowed to be `None`
    kernel_size: int, or sequence of int, default 1,
        kernel size (filter size) of the layer, should be compatible with `input_shape`
    stride: int, or sequence of int, default 1,
        stride (down-sampling length) of the layer, should be compatible with `input_shape`
    padding: int, or sequence of int, default 0,
        padding length(s) of the layer, should be compatible with `input_shape`
    channel_last: bool, default False,
        channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention)

    Returns
    -------
    output_shape: tuple,
        shape of the output Tensor

    """
    output_shape = compute_output_shape(
        "avgpool",
        input_shape,
        1,
        kernel_size,
        stride,
        padding,
        0,
        1,
        channel_last,
    )
    return output_shape


def compute_deconv_output_shape(
    input_shape: Sequence[Union[int, None]],
    num_filters: Optional[int] = None,
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    output_padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """

    compute the output shape of a transpose convolution layer

    Parameters
    ----------
    input_shape: sequence of int or None,
        shape of an input Tensor,
        the first dimension is the batch dimension, which is allowed to be `None`
    num_filters: int, optional,
        number of filters, also the channel dimension
    kernel_size: int, or sequence of int, default 1,
        kernel size (filter size) of the layer, should be compatible with `input_shape`
    stride: int, or sequence of int, default 1,
        stride (down-sampling length) of the layer, should be compatible with `input_shape`
    padding: int, or sequence of int, default 0,
        padding length(s) of the layer, should be compatible with `input_shape`
    out_padding: int, or sequence of int, default 0,
        additional size added to one side of the output shape,
        used only for transpose convolution
    dilation: int, or sequence of int, default 1,
        dilation of the layer, should be compatible with `input_shape`
    channel_last: bool, default False,
        channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention)

    Returns
    -------
    output_shape: tuple,
        shape of the output Tensor

    """
    output_shape = compute_output_shape(
        "deconv",
        input_shape,
        num_filters,
        kernel_size,
        stride,
        padding,
        output_padding,
        dilation,
        channel_last,
    )
    return output_shape


compute_sequential_output_shape_docstring = """

    Parameters
    ----------
    seq_len: int,
        length of the 1d sequence
    batch_size: int, optional,
        the batch size, can be None

    Returns
    -------
    output_shape: sequence,
        the output shape, given `seq_len` and `batch_size`

    """


def compute_sequential_output_shape(
    model: nn.Sequential,
    seq_len: Optional[int] = None,
    batch_size: Optional[int] = None,
) -> Sequence[Union[int, None]]:
    """

    compute the output shape of a sequential model

    Parameters
    ----------
    model: nn.Sequential,
        the sequential model
    seq_len: int,
        length of the 1d sequence
    batch_size: int, optional,
        the batch size, can be None

    Returns
    -------
    output_shape: sequence,
        the output shape, given `seq_len` and `batch_size`

    """
    assert issubclass(
        type(model), nn.Sequential
    ), f"model should be nn.Sequential, but got {type(model)}"
    _seq_len = seq_len
    for module in model:
        output_shape = module.compute_output_shape(_seq_len, batch_size)
        _, _, _seq_len = output_shape
    return output_shape


def compute_module_size(
    module: nn.Module, human: bool = False, dtype: str = "float32"
) -> Union[int, str]:
    """

    compute the size (number of parameters) of a module

    Parameters
    ----------
    module: Module,
        a torch Module
    human: bool, default False,
        return size in a way that is easy to read by a human,
        by appending a suffix corresponding to the unit (K, M, G, T, P)
    dtype: str, default "float32",
        data type of the module parameters, one of "float16", "float32", "float64"

    Returns
    -------
    n_params: int,
        size (number of parameters) of this torch module

    """
    module_parameters = filter(lambda p: p.requires_grad, module.parameters())
    n_params = sum([np.prod(p.size()) for p in module_parameters])
    if human:
        n_params = (
            n_params * {"float16": 2, "float32": 4, "float64": 8}[dtype.lower()] / 1024
        )
        div_count = 0
        while n_params >= 1024:
            n_params /= 1024
            div_count += 1
        # cvt_dict = {0:"K", 1:"M", 2:"G", 3:"T", 4:"P"}
        cvt_dict = {c: u for c, u in enumerate(list("KMGTP"))}
        n_params = f"""{n_params:.1f}{cvt_dict[div_count]}"""
    return n_params


def compute_receptive_field(
    kernel_sizes: Union[Sequence[int], int] = 1,
    strides: Union[Sequence[int], int] = 1,
    dilations: Union[Sequence[int], int] = 1,
    input_len: Optional[int] = None,
) -> Union[int, float]:
    r"""

    computes the (generic) receptive field of feature map of certain channel,
    from certain flow (if not merged, different flows, e.g. shortcut, must be computed separately),
    for convolutions, (non-global) poolings.
    "generic" refers to a general position, rather than specific positions,
    like on the edges, whose receptive field is definitely different

    .. math::
        Let the layers has kernel size, stride, dilation $(k_n, s_n, d_n)$ respectively.
        Let each feature map has receptive field length $r_n$,
        and difference of receptive fields of adjacent positions be $f_n$.
        By convention, $(r_0, f_0) = (1, 1)$. Then one has
        \begin{eqnarray}
        r_{n+1} & = & r_n + d_n(k_n-1)f_n, \\
        f_{n+1} & = & s_n f_n.
        \end{eqnarray}
        Hence
        \begin{eqnarray}
        f_{n} & = & \prod\limits_{i=0}^{n-1} s_i, \\
        r_{n} & = & 1 + \sum\limits_{i=0}^{n-1} d_i(k_i-1) \prod\limits_{i=0}^{j-1} s_j,
        \end{eqnarray}
        with empty products equaling 1 by convention.

    Parameters
    ----------
    kernel_sizes: int or sequence of int,
        the sequence of kernel size for all the layers in the flow
    strides: int or sequence of int, default 1,
        the sequence of strides for all the layers in the flow
    dilations: int or sequence of int, default 1,
        the sequence of strides for all the layers in the flow
    input_len: int, optional,
        length of the first feature map in the flow

    Returns
    -------
    receptive_field: int,
        (length) of the receptive field

    Examples
    --------
    >>> compute_receptive_field([11,2,7,7,2,5,5,5,2],[1,2,1,1,2,1,1,1,2])
    90
    >>> compute_receptive_field([11,2,7,7,2,5,5,5,2],[1,2,1,1,2,1,1,1,2],[2,1,2,4,1,8,8,8,1])
    484
    >>> compute_receptive_field([11,2,7,7,2,5,5,5,2],[1,2,1,1,2,1,1,1,2],[4,1,4,8,1,16,32,64,1])
    1984

    this is the receptive fields of the output feature maps
    of the 3 branches of the multi-scopic net, using its original hyper-parameters,
    (note the 3 max pooling layers)

    """
    _kernel_sizes = (
        [kernel_sizes] if isinstance(kernel_sizes, int) else list(kernel_sizes)
    )
    num_layers = len(_kernel_sizes)
    _strides = (
        list(repeat(strides, num_layers)) if isinstance(strides, int) else list(strides)
    )
    _dilations = (
        list(repeat(dilations, num_layers))
        if isinstance(dilations, int)
        else list(dilations)
    )
    assert num_layers == len(_strides) == len(_dilations)
    receptive_field = 1
    for idx, (k, d) in enumerate(zip(_kernel_sizes, _dilations)):
        s = np.prod(_strides[:idx]) if idx > 0 else 1
        receptive_field += d * (k - 1) * s
    if input_len is not None:
        receptive_field = min(receptive_field, input_len)
    return receptive_field


def default_collate_fn(
    batch: Sequence[Union[Tuple[np.ndarray, ...], Dict[str, np.ndarray]]]
) -> Union[Tuple[Tensor, ...], Dict[str, Tensor]]:
    """
    collate functions for model training

    the data generator (`Dataset`) should generate (`__getitem__`) n-tuples `signals, labels, ...`,
    or dictionaries of tensors

    Parameters
    ----------
    batch: sequence,
        sequence of n-tuples,
        in which the first element is the signal, the second is the label, ...;
        or sequence of dictionaries of tensors

    Returns
    -------
    tuple or dict of Tensor,
        the concatenated values to feed into neural networks
    """
    if isinstance(batch[0], dict):
        keys = batch[0].keys()
        collated = _default_collate_fn([tuple(b[k] for k in keys) for b in batch])
        return {k: collated[i] for i, k in enumerate(keys)}
    else:
        return _default_collate_fn(batch)


def _default_collate_fn(batch: Sequence[Tuple[np.ndarray, ...]]) -> Tuple[Tensor, ...]:
    """
    collate functions for model training

    the data generator (`Dataset`) should generate (`__getitem__`) n-tuples `signals, labels, ...`

    Parameters
    ----------
    batch: sequence,
        sequence of n-tuples,
        in which the first element is the signal, the second is the label, ...

    Returns
    -------
    tuple of Tensor,
        the concatenated values to feed into neural networks

    """
    try:
        n_fields = len(batch[0])
    except Exception:
        raise ValueError("No data")
    ret = []
    for i in range(n_fields):
        values = [[item[i]] for item in batch]
        values = np.concatenate(values, axis=0)
        values = torch.from_numpy(values)
        ret.append(values)
    return tuple(ret)


if torch.__version__ >= "1.15.0":

    def _true_divide(dividend, divisor):
        return torch.true_divide(dividend, divisor)

else:

    def _true_divide(dividend, divisor):
        return dividend / divisor


def _adjust_cnn_filter_lengths(
    config: dict,
    fs: int,
    ensure_odd: bool = True,
    pattern: str = "filter_length|filter_size",
) -> dict:
    """
    adjust the filter lengths (kernel sizes) in the config for convolutional neural networks,
    according to the new sampling frequency

    Parameters
    ----------
    config: dict,
        the config dictionary,
        this dict is NOT modified
    fs: int,
        the new sampling frequency
    ensure_odd: bool, default True,
        if True, the new filter lengths are ensured to be odd
    pattern: str, default "filter_length|filter_size",
        the pattern to search for in the config items related to filter lengths

    Returns
    -------
    config: dict,
        the adjusted config dictionary

    """
    assert "fs" in config
    config = deepcopy(config)
    for k, v in config.items():
        if isinstance(v, dict):
            tmp_config = deepcopy(v)
            original_fs = tmp_config.get("fs", None)
            tmp_config.update({"fs": config["fs"]})
            config[k] = _adjust_cnn_filter_lengths(tmp_config, fs, ensure_odd, pattern)
            if original_fs is None:
                config[k].pop("fs", None)
            else:
                config[k]["fs"] = fs
        elif re.findall(pattern, k):
            if isinstance(v, (Sequence, np.ndarray)):  # DO NOT use `Iterable`
                config[k] = [
                    _adjust_cnn_filter_lengths(
                        {"filter_length": fl, "fs": config["fs"]}, fs, ensure_odd
                    )["filter_length"]
                    for fl in v
                ]
            elif isinstance(v, Real):
                # DO NOT use `int`, which might not work for numpy array elements
                if v > 1:
                    config[k] = int(round(v * fs / config["fs"]))
                    if ensure_odd:
                        config[k] = config[k] - config[k] % 2 + 1
        elif isinstance(v, Sequence) and not isinstance(v, (str, bytes)):
            tmp_configs = [
                _adjust_cnn_filter_lengths(
                    {k: item, "fs": config["fs"]}, fs, ensure_odd, pattern
                )
                for item in v
            ]
            config[k] = [item[k] for item in tmp_configs]
    return config


def adjust_cnn_filter_lengths(
    config: dict,
    fs: int,
    ensure_odd: bool = True,
    pattern: str = "filter_length|filter_size",
) -> dict:
    """
    adjust the filter lengths in the config for convolutional neural networks,
    according to the new sampling frequency

    Parameters
    ----------
    config: dict,
        the config dictionary,
        this dict is NOT modified
    fs: int,
        the new sampling frequency
    ensure_odd: bool, default True,
        if True, the new filter lengths are ensured to be odd
    pattern: str, default "filter_length|filter_size",
        the pattern to search for in the config items related to filter lengths

    Returns
    -------
    config: dict,
        the adjusted config dictionary

    """
    config = _adjust_cnn_filter_lengths(config, fs, ensure_odd, pattern)
    config["fs"] = fs
    return config


class SizeMixin(object):
    """Mixin class for size related methods"""

    @property
    def module_size(self) -> int:
        return compute_module_size(
            self,
            dtype=self.dtype_,
        )

    @property
    def module_size_(self) -> str:
        return compute_module_size(
            self,
            human=True,
            dtype=self.dtype_,
        )

    @property
    def dtype(self) -> torch.dtype:
        try:
            return next(self.parameters()).dtype
        except StopIteration:
            return torch.float32
        except Exception as err:
            raise err  # unknown error

    @property
    def device(self) -> torch.device:
        try:
            return next(self.parameters()).device
        except StopIteration:
            return torch.device("cpu")
        except Exception as err:
            raise err  # unknown error

    @property
    def dtype_(self) -> str:
        return str(self.dtype).replace("torch.", "")

    @property
    def device_(self) -> str:
        return str(self.device)


class CkptMixin(object):
    """Mixin class for loading from checkpoint class methods"""

    @classmethod
    def from_checkpoint(
        cls, path: str, device: Optional[torch.device] = None
    ) -> Tuple[nn.Module, dict]:
        """

        load a model from a checkpoint

        Parameters
        ----------
        path: str,
            path of the checkpoint
        device: torch.device, optional,
            map location of the model parameters,
            defaults "cuda" if available, otherwise "cpu"

        Returns
        -------
        model: Module,
            the model loaded from a checkpoint
        aux_config: dict,
            auxiliary configs that are needed for data preprocessing, etc.

        """
        _device = device or DEFAULTS.device
        ckpt = torch.load(path, map_location=_device)
        aux_config = ckpt.get("train_config", None) or ckpt.get("config", None)
        assert (
            aux_config is not None
        ), "input checkpoint has no sufficient data to recover a model"
        kwargs = dict(
            classes=aux_config["classes"],
            config=ckpt["model_config"],
        )
        if "n_leads" in aux_config:
            kwargs["n_leads"] = aux_config["n_leads"]
        model = cls(**kwargs)
        model.load_state_dict(ckpt["model_state_dict"])
        return model, aux_config
